/*
    Copyright (C) 2019 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#include "EspNowSlave.h"
#include <ESP8266WiFi.h>
#include "mavlink/common/mavlink.h"
extern "C" {
  #include <espnow.h>
  #include "user_interface.h"
}
#include "utils.h"

namespace NanopixCore {

EspNowSlave * EspNowSlave::_instance = NULL;

void espNowSentCallback(uint8_t* mac, uint8_t sendStatus);
void espNowReceivedCallback(uint8_t * mac, uint8_t * data, uint8_t len);

EspNowSlave::EspNowSlave()
{ 
  _packetReceivedCallback = NULL;
  _instance = NULL;
}

EspNowSlave * EspNowSlave::getInstance()
{
  if(_instance == NULL)
  {
    _instance = new EspNowSlave();
  }
  return _instance;
}

bool EspNowSlave::initializeConnection()
{
  WiFi.softAPdisconnect(true);
  TRACE_PRINTF("EspNowSlave::initializeConnection: softAPdisconnect(true) completed\n");
  WiFi.mode(WIFI_STA); // Station mode for esp-now controller
  TRACE_PRINTF("EspNowSlave::initializeConnection: mode(WIFI_STA) completed\n");
  if (esp_now_init() != 0) 
  {
      //_debug_serial->println("EspNowManager startEspNow: esp_now_init failed, restart...");
      delay(100);
      ESP.restart();
  }
  esp_now_set_self_role(ESP_NOW_ROLE_CONTROLLER);
  
  esp_now_register_send_cb(espNowSentCallback);
  esp_now_register_recv_cb(espNowReceivedCallback);

  uint8_t heartbeat_reply[200];
  uint16_t heartbeat_len;
  mavlink_message_t msg;
  mavlink_msg_heartbeat_pack_chan(0, 0, 0, &msg, MAV_TYPE_QUADROTOR, MAV_AUTOPILOT_GENERIC, 0, 0, MAV_STATE_BOOT);
  heartbeat_len = mavlink_msg_to_send_buffer(heartbeat_reply, &msg);

  int stop_pairing = millis() + PAIRING_STAGE_MS;
  int sent_count = 0;
  while(!_pairing_reply_received && (millis() < stop_pairing))
  {
      esp_now_send(_server_mac, heartbeat_reply, heartbeat_len);
      ++sent_count;
      delay(1000 / PAIRING_STAGE_FREQ_HZ);
  }
  return _pairing_reply_received;
  /*_debug_serial = new SoftwareSerial(SW_SERIAL_UNUSED_PIN, 5);
  _debug_serial->begin(38400);
  if(espnow_reply_received)
  {
    _debug_serial->print("EspNow CONNECTED! Sent count: ");
    _debug_serial->println(sent_count);
  }
  else
  {
    _debug_serial->println("EspNow connect TIMEOUT!");
  }*/
}

void espNowSentCallback(uint8_t* mac, uint8_t sendStatus) 
{
  EspNowSlave::getInstance()->_msg_sent = true;
}

void espNowReceivedCallback(uint8_t * mac, uint8_t * data, uint8_t len) 
{
  auto inst = EspNowSlave::getInstance();
  if(!esp_now_is_peer_exist(mac))
  {
    esp_now_add_peer(mac, ESP_NOW_ROLE_SLAVE, 1, NULL, 0); // 'NULL, 0' for disable encryption
    if(mac != NULL)
    {
      memcpy(inst->_controller_mac, mac, 6);
    }
    inst->_pairing_reply_received = true;
  }
  else
  {
    if(inst->_packetReceivedCallback != NULL)
    {
      (*(inst->_packetReceivedCallback))(data, len);
    }
  }
}

void EspNowSlave::setOnReceivedCallback(void (*packetReceivedCallback)(uint8_t * data, uint8_t len))
{
  _packetReceivedCallback = packetReceivedCallback;
}

bool EspNowSlave::_trySend(uint8_t * data, size_t length)
{
  if(_msg_sent || (millis() > _last_msg_sent_ts + MAX_DELIVERY_TIME_MS))
  {
    _msg_sent = false;
    _last_msg_sent_ts = millis();
    esp_now_send(_controller_mac, data, length);
    return true;
  } else {
    return false;
  }
}

void EspNowSlave::addDataToSend(uint8_t * data, size_t length)
{
  bool try_send_res = false;
  if(_packets_tx_fifo.count() == 0)
  {
    try_send_res = _trySend(data, length);
  }
  if(!try_send_res)
  {
    _packets_tx_fifo.pushPacket(data, length);
  }
}

void EspNowSlave::flush()
{
  if(_msg_sent || (millis() > _last_msg_sent_ts + MAX_DELIVERY_TIME_MS))
  {
    uint8_t * data = NULL;
    size_t len = 0;
    if(_packets_tx_fifo.pullPacket(&data, len))
    {
      if(data != NULL)
      {
        _trySend(data, len);
      }
    }
  }
}

}