/*
    Copyright (C) 2019 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#include "EspOTA.h"
#include "nanopix_config.h"
#include "utils.h"
#include "string.h"
#include "stm_boot_manager.h"
#include "Updater.h"


namespace NanopixCore {

EspOTA::EspOTA()
{
    _resetState();
}

void EspOTA::_resetState()
{
    _esp_ota_progress = 0;
    _esp_ota_fw_size = 0;
    memset(_esp_firmware_md5, 0, FIRMWARE_MD5_BUFFER_LEN);
    _esp_ota_next_addr = 0;
    _esp_ota_success = false;
}

mavlink_log_data_t _file_msg; // do not move to class members
mavlink_message_t _ota_tx_mav_msg;

int EspOTA::handleEspOtaPacket(mavlink_message_t * mavmsg, mavlink_message_t ** out_responce_msg)
{
  *out_responce_msg = NULL;

  uint32_t addr = 0;
  uint8_t data_size = 0;
  bool is_size_accepted = true;
  int progress = 0;
  uint8_t * update_data = NULL;
  int res_error_code = 0;

  mavlink_msg_log_data_decode(mavmsg, &_file_msg);

  addr = _file_msg.ofs;
  data_size = _file_msg.count;
  if(addr == 0)
  {
    _resetState();
    update_data = _file_msg.data + (MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN - data_size);
    _esp_ota_fw_size = *((uint32_t*)(_file_msg.data));
    if(_esp_ota_fw_size > 0)
    {
      Update.end();
      is_size_accepted = Update.begin(_esp_ota_fw_size);
      if(is_size_accepted)
      {
        STM_BootManager::toResetState();
        Update.runAsync(true);
        _esp_firmware_md5[FIRMWARE_MD5_BUFFER_LEN - 1] = 0;
        memcpy(_esp_firmware_md5, _file_msg.data + 4, FIRMWARE_MD5_LEN);
        Update.setMD5(_esp_firmware_md5);
        
        DEBUG_PRINTF("Expected MD5: %s\n", _esp_firmware_md5);
      } else {
        DEBUG_PRINTF("ESP OTA ERROR, not enougth space: %d\n", (int)_esp_ota_fw_size);
        res_error_code = -3;
      }
    }
    else
    {
      is_size_accepted = false;
      DEBUG_PRINTF("ESP OTA ERROR, fw size = 0\n");
      res_error_code = -2;
    }
    
  } else {
    update_data = _file_msg.data;
  }

  if(is_size_accepted && (addr == _esp_ota_next_addr) && (addr < _esp_ota_fw_size))
  {
    if((update_data != NULL) && (data_size != 0))
    {
      Update.write(update_data, data_size);
    }
    _esp_ota_next_addr += data_size;
    progress = (int)(100 * (Update.progress() / (float)_esp_ota_fw_size));
    if(progress != _esp_ota_progress)
    {
      _esp_ota_progress = progress;
      DEBUG_PRINTF("Progress: %d\n", _esp_ota_progress);
    }

    if(Update.isFinished())
    {
      _esp_ota_success = Update.end();
      if(!_esp_ota_success)
      {
        res_error_code = (int)Update.getError();
        DEBUG_PRINTF("ESP OTA error code: %d\n", res_error_code);
        // TODO reply ESP OTA error code
      }
    }
  }

  if((addr == 0) || (addr > _esp_ota_next_addr) || addr == _esp_ota_fw_size)
  {
    int request_data_size = 90;
    if(addr == _esp_ota_fw_size && data_size == 0)
    {
      request_data_size = _esp_ota_success;
    }
    if(addr == _esp_ota_fw_size && data_size == 1)
    {
      DEBUG_PRINTF("Updated successfully! Restarting...\n");
      delay(40);
      ESP.restart();
      return res_error_code;
    }
    if(addr == _esp_ota_fw_size && data_size == 2)
    {
      DEBUG_PRINTF("Flashing cancelled! Exit.\n");
      _resetState();
      return -1;
    }
    //DEBUG_PRINTF("ESP OTA expected addr %d, got %d\n", (int)_esp_ota_next_addr, (int)addr);
    if(is_size_accepted)
      mavlink_msg_log_request_data_pack_chan(0, 0, 1, &_ota_tx_mav_msg, 0, 0, 0, 
                                             _esp_ota_next_addr, request_data_size);
    else
      mavlink_msg_log_request_data_pack_chan(0, 0, 1, &_ota_tx_mav_msg, 0, 0, 0, 0, 0);
    
    *out_responce_msg = &_ota_tx_mav_msg;
  }
  return res_error_code;
}

}