/*
    Copyright (C) 2020 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#include "EspUpdater.h"
#include "Updater.h"
#include "stdlib.h"
#include "utils.h"
#include "nanopix_config.h"

namespace NanopixCore {

EspUpdater * EspUpdater::_instance = NULL;

EspUpdater::EspUpdater()
{
  _temp_buffer = NULL;
  _buffer_processed_len = 0;
  _retry_write_counter = 0;
  _custom_fatal_code = 0;
  _status = IUpdater::NOT_STARTED;
}

EspUpdater & EspUpdater::getInstance()
{
    if(_instance == NULL)
    {
        _instance = new EspUpdater();
    }
    return *_instance;
}

bool EspUpdater::begin(uint32_t firmware_size, uint8_t * md5_bytes)
{
  reset();
  bool is_size_valid = Update.begin(firmware_size);
  if(is_size_valid)
  {
    Update.runAsync(false);
    char md5_hex_string[33];
    md5BytesToHexString(md5_bytes, md5_hex_string);
    md5_hex_string[32] = 0;
    Update.setMD5(md5_hex_string);
    _retry_write_counter = 0;
    _custom_fatal_code = 0;
    _status = IUpdater::IN_PROCESS;
  } else {
    _status = IUpdater::NOT_STARTED;
  }
  return is_size_valid;
}

void EspUpdater::md5BytesToHexString(uint8_t * input_bytes, char * output_string) // input 16 bytes, output 32 chars
{
    for(uint8_t i = 0; i < 16; ++i) 
    {
        sprintf(output_string + (i * 2), "%02x", input_bytes[i]);
    }
}

IUpdater::UpdateStatus EspUpdater::getStatus() const
{
  return _status;
}

void EspUpdater::end()
{
  _temp_buffer = NULL;
  bool md5_ok = Update.end();
  if(md5_ok)
  {
    _status = COMPLETED;
  }
  else
  {
    _status = FAILED;
  }
}

bool EspUpdater::isFwSizeValid(int32_t firmware_size)
{
  if(firmware_size <= 0 || firmware_size > MAX_ESP_FIRMWARE_SIZE)
    return false;
  Update.end();
  return Update.begin(firmware_size);
}

bool EspUpdater::isBufferCompleted()
{
  if(_temp_buffer == NULL)
    return true;
  return (_buffer_processed_len == _temp_buffer->length);
}

void EspUpdater::setBuffer(FwBuffer * temp_buffer)
{
  _temp_buffer = temp_buffer;
  _buffer_processed_len = 0;
}

void EspUpdater::reset()
{
  _status = IUpdater::NOT_STARTED;
  Update.end();
  Update.clearError();
  _buffer_processed_len = 0;
  _temp_buffer = NULL;
  _retry_write_counter = 0;
  _custom_fatal_code = 0;
}

int EspUpdater::getLastFatalCode()
{
  return _custom_fatal_code;
}

IUpdater::HandleResult EspUpdater::getLastHandleResult()
{
  return _last_handle_result;
}

IUpdater::HandleResult EspUpdater::handle()
{
  if(_status == IN_PROCESS) 
  {
    size_t remaining_len = 0;
    size_t completed_len = 0;
    if(_temp_buffer != NULL)
    {
      if(_buffer_processed_len < _temp_buffer->length)
      {
        remaining_len = _temp_buffer->length - _buffer_processed_len;
        completed_len = Update.write(&(_temp_buffer->data[_buffer_processed_len]), remaining_len);
        _buffer_processed_len += completed_len;
        
      }
      else
      {
        _last_handle_result = HandleResult::NO_NEW_DATA;
      }
    }

    if(Update.isFinished())
    {
      _status = FINISHED;
    }
    
    _custom_fatal_code = Update.getError();
    if(_custom_fatal_code != 0)
    {
      if(_retry_write_counter < MAX_RETRY_WRITE_COUNT)
      {
        ++_retry_write_counter;
        _last_handle_result = HandleResult::ERROR_RETRY_BLOCK;
      }
      else
      {
        _last_handle_result = HandleResult::FATAL_ERROR;
      }
    }
    else
    {
      _retry_write_counter = 0;
      if(remaining_len == completed_len)
      {
        _last_handle_result = HandleResult::HANDLE_COMPLETED;
      }
      else
      {
        _last_handle_result = HandleResult::WAITING;
      }
    }
  }
  return _last_handle_result;
}

}
