/*
    Copyright (C) 2020 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/
#include "FirmwareDataReceiver.h"
#include "platform_hal.h"

namespace NanopixCore {

FirmwareDataReceiver * FirmwareDataReceiver::_instance = NULL;
mavlink_log_data_t _fw_log_msg; // do not move to class members
mavlink_message_t _tx_mav_msg;
uint8_t _buffer1_data[FirmwareDataReceiver::BUFFER_SIZE];
uint8_t _buffer2_data[FirmwareDataReceiver::BUFFER_SIZE];

FirmwareDataReceiver::FirmwareDataReceiver()
{
  _initialization_finished = false;

  _buffer1.data = _buffer1_data;
  _buffer1.size = FirmwareDataReceiver::BUFFER_SIZE;

  _buffer2.data = _buffer2_data;
  _buffer2.size = FirmwareDataReceiver::BUFFER_SIZE;

  for(int i = 0; i < MAX_SYSTEMS_COUNT; ++i)
  {
    _updaters[i] = NULL;
  }

  _reset();

  _last_packet_ts = 0;
}

void FirmwareDataReceiver::_reset()
{
  _is_last_frame_received = false;
  _fw_frames_count = 0;
  _firmware_size = 0;
  _pack_count_in_frame = 0;
  _update_for_system_id = 0;
  _current_frame_id = 0;
  _buffer1.length = 0;
  _buffer2.length = 0;
  _active_buffer = &_buffer1;
  memset(_firmware_md5, 0, FIRMWARE_MD5_LEN);
  memset(_frame_md5, 0, FIRMWARE_MD5_LEN);
  memset(_start_response_data, 0, MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN);
  _start_response_data[0] = PROTOCOL_VERSION;
  _frame_fill_data.resetAll();
  _md5.begin();
}

void FirmwareDataReceiver::handlePacket(mavlink_message_t * mavmsg,
                                        mavlink_message_t ** out_response_msg)
{
  
  bool is_need_response = false;
  bool is_frame_completed = false;
  int resp_code = (int)NO_ERRORS;
  int input_packet_frame_index = 0;
  bool is_start_pack = false;
  if(_initialization_finished)
  {
    if(systemMillis() - _last_packet_ts > PACKET_TIMEOUT_MS)
    {
      if(_updaters[_update_for_system_id] != NULL)
      {
          _updaters[_update_for_system_id]->reset();
      }
    }

    ////////////////////////// Decode and parse input data packet /////////////////////////////
    mavlink_msg_log_data_decode(mavmsg, &_fw_log_msg);
    is_start_pack = (_fw_log_msg.ofs == 0 && _fw_log_msg.id == 0 && _fw_log_msg.count == 0);
    input_packet_frame_index = _fw_log_msg.ofs;
    if(is_start_pack)
    {
      resp_code = _processStartPacket(&_fw_log_msg);
      if(resp_code < 128)
      {
        _updaters[_update_for_system_id]->begin(_firmware_size, (uint8_t*)(&(_firmware_md5[0])));
      }
      is_need_response = true;
    }
    else
    {
      if(_updaters[_update_for_system_id] != NULL)
      {
          if(_updaters[_update_for_system_id]->getStatus() != IUpdater::NOT_STARTED)
          {
            resp_code = _processRegularPacket(&_fw_log_msg, is_need_response, is_frame_completed);
          }
          else
          {
            is_need_response = true;
            resp_code = NEED_RESTART;
          }
      }
      else
      {
          is_need_response = true;
          resp_code = UNSUPPORTED_SYSTEM_ID;
      }
    }

    if(_updaters[_update_for_system_id] && 
      _updaters[_update_for_system_id]->getLastHandleResult() == IUpdater::FATAL_ERROR)
    {
      resp_code = (int)CUSTOM_FATAL_CODE;
      is_need_response = true;
    }
  }
  else 
  {
    is_frame_completed = false;
    _current_frame_id = 0;
    input_packet_frame_index = 0;
    resp_code = SERVICE_NOT_INITIALIZED_YET;
    is_need_response = true;
  }

  ////////////////////////// Send response mavlink packet ///////////////////////////////////
  if(is_need_response)
  {
    uint32_t received_packs_in_frame = _frame_fill_data.getSetCount();
    uint16_t request_packs = (uint16_t)(_pack_count_in_frame - received_packs_in_frame);
    if(input_packet_frame_index < _current_frame_id)  // all is ok for prev completed packet
    {
      request_packs = 0;
      resp_code = NO_ERRORS;
    }
    if(resp_code == CUSTOM_FATAL_CODE)
    {
      if(_updaters[_update_for_system_id] != NULL)
      {
        request_packs = _updaters[_update_for_system_id]->getLastFatalCode();
      }
      else
      {
        request_packs = -1;
      }
    }

    mavlink_msg_log_data_pack_chan(0, 0, 1, &_tx_mav_msg, request_packs, input_packet_frame_index, 
                                   resp_code, is_start_pack ? _start_response_data : _frame_fill_data.mask);
    *out_response_msg = &_tx_mav_msg;
  }
  else
  {
    *out_response_msg = NULL;
  }

  ////////////////////////// Reset if frame completed ////////////////////////////////////////
  if(is_frame_completed)
  {
    _frame_fill_data.resetAll();
  }

  _last_packet_ts = systemMillis();
}

int FirmwareDataReceiver::_processRegularPacket(mavlink_log_data_t * packet, 
                                                bool & out_need_response,
                                                bool & out_frame_completed)
{
  int res = (int)NO_ERRORS;
  out_frame_completed = false;
  int frame_index = packet->ofs;
  uint32_t frame_packet_index = packet->id;
  bool is_last_pack = (frame_packet_index == _pack_count_in_frame - 1);
  out_need_response = is_last_pack;
  if(frame_index == _current_frame_id)
  {
    uint32_t buffer_pos = frame_packet_index * MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN;
    if(buffer_pos + MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN <= BUFFER_SIZE)
    {
      if(!_frame_fill_data.isBitSet(frame_packet_index)) 
      {
        if(packet->count > 0) 
        {
          memcpy(&(_active_buffer->data[buffer_pos]), packet->data, packet->count);
          _active_buffer->length += packet->count;
        }
        _frame_fill_data.setBit(frame_packet_index);
      }
      if(is_last_pack)
      {
        uint32_t received_packs_in_frame = _frame_fill_data.getSetCount();
        if(received_packs_in_frame == _pack_count_in_frame)
        {
          if(_updaters[_update_for_system_id]->isBufferCompleted())
          {
            // last packet in frame contains frame md5 at the end
            memcpy(_frame_md5, &(packet->data[MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN - FIRMWARE_MD5_LEN]), 
                   FIRMWARE_MD5_LEN);
            _md5.add(_active_buffer->data, _active_buffer->length);
            _md5.calculate();
            uint8_t md5_bytes[FIRMWARE_MD5_LEN];
            _md5.getBytes(md5_bytes);
            if(memcmp(md5_bytes, _frame_md5, FIRMWARE_MD5_LEN) == 0 || _is_last_frame_received)
            {
              if(_current_frame_id == _fw_frames_count - 1)
              {
                _is_last_frame_received = true;
                IUpdater::UpdateStatus stat = _updaters[_update_for_system_id]->getStatus();
                if(stat == IUpdater::UpdateStatus::IN_PROCESS)
                {
                  res = (int)WAIT_FLASHING_PROCESS;
                }
                if(stat == IUpdater::UpdateStatus::FINISHED)
                {
                  out_frame_completed = true;
                  _updaters[_update_for_system_id]->end();
                  stat = _updaters[_update_for_system_id]->getStatus();
                  if(stat == IUpdater::UpdateStatus::FAILED) 
                  {
                    res = (int)RESULT_MD5_FAILED;
                  }
                }
              }
              else
              {
                ++_current_frame_id;
                out_frame_completed = true;
              }
              _updaters[_update_for_system_id]->setBuffer(_switchBuffer());
            } 
            else 
            {
              _frame_fill_data.resetAll();
              _active_buffer->length = 0;
              res = (int)INVALID_FRAME_MD5;
            }
            _md5.begin();
          } 
          else
          {
            res = (int)WAIT_FLASHING_PROCESS;
          }
        }
      }
    }
    else
    {
      res = (int)FRAME_PACKET_OUT_OF_BUFFER;
    }
  }
  else
  {
    if(!is_last_pack) 
    {
      out_need_response = true;
      res = (int)INVALID_FRAME_ID;
    }
  }
  return res;
}

void FirmwareDataReceiver::finishInitialization()
{
  _initialization_finished = true;
}

FwBuffer * FirmwareDataReceiver::_switchBuffer()
{
  if(_active_buffer == &_buffer1)
  {
    _active_buffer = &_buffer2;
    _active_buffer->length = 0;
    return &_buffer1;
  } 
  else
  {
    _active_buffer = &_buffer1;
    _active_buffer->length = 0;
    return &_buffer2;
  }
}

int FirmwareDataReceiver::_processStartPacket(mavlink_log_data_t * packet)
{
  // start packet data map:
  // [0..15] firmware binary md5 hashsum
  // [16..19] uint32_t firmware size (bytes)
  // [20..23] uint32_t data frame size (bytes)
  // [24] update for system id
  _reset();
  int resp_code = (int)NO_ERRORS;
  memcpy(_firmware_md5, packet->data, FIRMWARE_MD5_LEN);

  memcpy((void*)(&_firmware_size), (void*)(&(packet->data[16])), 4);
  
  int32_t pref_frame_size;
  memcpy((void*)(&pref_frame_size), (void*)(&(packet->data[20])), 4);

  memcpy((void*)(&_fw_frames_count), (void*)(&(packet->data[24])), 4);

  _update_for_system_id = packet->data[28];

  if(_update_for_system_id < MAX_SYSTEMS_COUNT)
  {
    if(_updaters[_update_for_system_id] == NULL)
    {
      return (int)UNSUPPORTED_SYSTEM_ID;
    }
  }
  else
  {
    return (int)UNSUPPORTED_SYSTEM_ID;
  }

  if(pref_frame_size > BUFFER_SIZE) 
  {
    pref_frame_size = BUFFER_SIZE;
  }
  _pack_count_in_frame = (pref_frame_size / MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN);
  if(pref_frame_size % MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN != 0)
  {
    resp_code = (int)INVALID_FRAME_SIZE;
  }
  if(_pack_count_in_frame == 0)
  {
    _pack_count_in_frame = BUFFER_SIZE / MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN;
    resp_code = (int)INVALID_FRAME_SIZE;
  }
  if(!(_updaters[_update_for_system_id]->isFwSizeValid(_firmware_size)))
  {
    resp_code = (int)INVALID_FW_SIZE;
  }
  //DEBUG_PRINTF("FW sender, fw size: %d, frame: %d\n", (int)_firmware_size, (int)pref_frame_size);
  return resp_code;
}

FirmwareDataReceiver::RegisteringResult FirmwareDataReceiver::registerSystemUpdater(
        uint8_t system_id, IUpdater * updater)
{
  if(!_initialization_finished) 
  {
    if(updater == NULL)
    {
      return NULL_UPDATER_REF;
    }
    if(system_id < MAX_SYSTEMS_COUNT)
    {
      if(_updaters[system_id] == NULL)
      {
        _updaters[system_id] = updater;
        return REGISTERED;
      }
      else
      {
        return ALREADY_EXISTS;
      }
      
    }
    else
    {
      return WRONG_SYSTEM_ID;
    }
  }
  else
  {
    return INIT_ALREADY_FINISHED;
  }
}

FirmwareDataReceiver & FirmwareDataReceiver::getInstance()
{
  if(_instance == NULL) 
  {
    _instance = new FirmwareDataReceiver();
  }
  return *_instance;
}

}

