/*
    Copyright (C) 2020 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#ifndef FIRMWARE_DATA_RECEIVER_H
#define FIRMWARE_DATA_RECEIVER_H

#include "stdint.h"
#include <MD5Builder.h>
#include "mavlink/common/mavlink.h"
#include "IUpdater.h"
#include "FwBuffer.h"

namespace NanopixCore {

class BitMask {
  uint32_t _set_counter;
public:
  BitMask()
  {
    _set_counter = 0;
  }

  uint8_t mask[MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN];
  
  void setBit(uint32_t bit_index)
  {
    if(bit_index < MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN * 8)
    {
      if(!_isSet(bit_index))
      {
        ++_set_counter;
      }
      mask[bit_index / 8] |= (1 << (bit_index % 8));
    }
  }

  bool isBitSet(uint32_t bit_index)
  {
    if(bit_index < MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN * 8)
    {
      return _isSet(bit_index);
    }
  }

  bool _isSet(uint32_t bit_index)
  {
    return mask[bit_index / 8] & ((uint8_t)(1 << (bit_index % 8)));
  }

  void resetAll()
  {
    memset(mask, 0, MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN);
    _set_counter = 0;
  }

  void setAll()
  {
    memset(mask, 0xFF, MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN);
    _set_counter = MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN * 8;
  }

  uint32_t getSetCount() const
  {
    return _set_counter;
  }
};

class FirmwareDataReceiver {
  static const uint8_t PROTOCOL_VERSION = 1;
  static const uint32_t PACKET_TIMEOUT_MS = 2000;
  static const int MAX_SYSTEMS_COUNT = 8;
  static const int FIRMWARE_MD5_LEN = 16;
  enum ResponseWarning { NO_ERRORS = 0, INVALID_FRAME_SIZE, INVALID_FRAME_MD5, WAIT_FLASHING_PROCESS };
  enum ResponseError { INVALID_FW_SIZE = 128, UNSUPPORTED_SYSTEM_ID, FRAME_PACKET_OUT_OF_BUFFER, 
                       INVALID_FRAME_ID, RESULT_MD5_FAILED, NEED_RESTART, CUSTOM_FATAL_CODE, SERVICE_NOT_INITIALIZED_YET };
public:
  enum RegisteringResult { REGISTERED = 0, ALREADY_EXISTS, WRONG_SYSTEM_ID, 
                           NULL_UPDATER_REF, INIT_ALREADY_FINISHED };
  static const size_t BUFFER_SIZE = 2700;

  void handlePacket(mavlink_message_t * mavmsg, mavlink_message_t ** out_response_msg);
  static FirmwareDataReceiver & getInstance();
  RegisteringResult registerSystemUpdater(uint8_t system_id, IUpdater * updater);
  void finishInitialization();
private:
  FirmwareDataReceiver();
  int _processStartPacket(mavlink_log_data_t * packet);
  int _processRegularPacket(mavlink_log_data_t * packet, 
                            bool & out_need_response, bool & out_frame_completed);
  void _reset();
  FwBuffer * _switchBuffer();

  static FirmwareDataReceiver * _instance;
  FwBuffer _buffer1, _buffer2;
  FwBuffer * _active_buffer;
  char _firmware_md5[FIRMWARE_MD5_LEN];
  char _frame_md5[FIRMWARE_MD5_LEN];
  int32_t _firmware_size;
  uint32_t _pack_count_in_frame;
  uint8_t _update_for_system_id;
  uint32_t _current_frame_id;
  int32_t _fw_frames_count;
  IUpdater * _updaters[MAX_SYSTEMS_COUNT];
  BitMask _frame_fill_data;
  uint32_t _last_packet_ts;
  bool _is_last_frame_received;
  uint8_t _start_response_data[MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN];
  volatile bool _initialization_finished;
  
  MD5Builder _md5;
};

}

#endif

