/*
    Copyright (C) 2017 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#include "NanopixCore.h"
#include "nanopix_config.h"
#include "utils.h"
#include "EspNowSlave.h"
#include "string.h"
#include "stm_boot_manager.h"
#include "FirmwareDataReceiver.h"
#include "EspUpdater.h"
#include "StmUpdater.h"
#include "STM32_OTA.h"
#include "Wire.h"

#define MAV_RX_CHAN 0
#define MAV_TX_CHAN 1

#define BOOT_MODE_WAIT_UPGRADE_MS (5 * 60 * 1000)

#include <SoftwareSerial.h>


//SoftwareSerial readerSerial(0, SW_SERIAL_UNUSED_PIN); // RX, TX

void internal_initialization() // called from <arduino>\hardware\esp8266_2.3.0\2.3.0\cores\esp8266\core_esp8266_main.cpp
{
#if DEBUG
  pinMode(STM32_BOOT_GPIO, OUTPUT);
  digitalWrite(STM32_BOOT_GPIO, LOW);
  NanopixCore::STM_BootManager::reset();
#endif

  WiFi.setOutputPower(20.5);

	NanopixCore::internal_init();
}

namespace NanopixCore {

int MainController::_claw_servo_ids[MAX_CLAW_COUNT] = {5, 6};

void espNowOnReceivedHandler(uint8_t * data, uint8_t len) 
{
  MainController::getInstance()._retransmitFromWireless(data, len);
}

void internal_init()
{
  /*SoftwareSerial _debug_serial(SW_SERIAL_UNUSED_PIN, 5);
  _debug_serial.begin(38400);
  softap_config	apconfig;
  if(wifi_softap_get_config(&apconfig))
  {
    if((apconfig.ssid[0] != 0) || (apconfig.password[0] != 0))
    {
      _debug_serial.print("AP config ssid: '");
      _debug_serial.print(std::string(reinterpret_cast< char const* >(apconfig.ssid)).c_str());
      _debug_serial.println("'");
      _debug_serial.print("AP config pass: '");
      _debug_serial.print(std::string(reinterpret_cast< char const* >(apconfig.password)).c_str());
      _debug_serial.println("'");
      //system_restore();
      //system_restart();
    }
  }*/
  
  //

  MainController & instance = MainController::getInstance();
	SettingsManager & settings_manager = SettingsManager::getInstance();
  settings_manager.initAll(&(instance.udpSender));
  SettingsManager::BOOT_MODES bootMode = settings_manager.getNextBootMode();
  settings_manager.setNextBootMode(SettingsManager::TRY_ESPNOW_BOOT_MODE);

  init_debug_serial();
  TRACE_PRINTF("Boot mode: %d\n", (int)bootMode);

  if(bootMode == SettingsManager::TRY_ESPNOW_BOOT_MODE)
  {
    instance._esp_now_mode = EspNowSlave::getInstance()->initializeConnection();
    TRACE_PRINTF("EspNow initializeConnection completed, res: %d\n", (int)instance._esp_now_mode);
    if(instance._esp_now_mode)
    {
      EspNowSlave::getInstance()->setOnReceivedCallback(espNowOnReceivedHandler);
    }
    else
    {
      settings_manager.setNextBootMode(SettingsManager::WIFI_BOOT_MODE);
      TRACE_PRINTF("Set wifi boot mode, restart\n");
      ESP.restart();
      return;
    }
  }

	instance.initializeSystem();
  TRACE_PRINTF("initializeSystem completed\n");
	
	if(bootMode == SettingsManager::UPGRADE_BOOT_MODE)
	{
		float blueColorIntens = 80;
		int blueColor = (int)blueColorIntens;
		int blueColorIntensStep = 6;
		
		instance.setFrontLEDColor(0, 0, 80);
		instance.setBackLEDColor(0, 0, 80);
		long startMillis = millis();
		long nextReportMs = millis();
		long remainingTimeMs = 0;
		while(remainingTimeMs >= 0)
		{
			remainingTimeMs = startMillis + BOOT_MODE_WAIT_UPGRADE_MS - millis();
			ArduinoOTA.handle();

			if(blueColor <= 15 || blueColor >= 150)
				blueColorIntensStep = -blueColorIntensStep;
			blueColor += blueColorIntensStep;

			instance.setFrontLEDColor(0, 0, blueColor);
			instance.setBackLEDColor(0, 0, blueColor);
			
			settings_manager.parseMavPackets();
			
			if(millis() > nextReportMs)
			{
				char dropLog[64];
				sprintf(dropLog, "ESP in boot mode, will reboot in %d sec", (int)(remainingTimeMs / 1000));
				instance.sendLogMsg(dropLog);
				nextReportMs += 1000;
			}
			
			delay(30);
		}
		
		instance.setFrontLEDColor(255, 0, 0);
		instance.setBackLEDColor(0, 255, 0);
	}
  instance._next_report_millis = millis() + 1000 / REPORT_FREQ_HZ;
}

MainController * mainControllerPtr = NULL;
os_timer_t stmToAppRetransmitTimer;
uint8_t prevPacketSeq = 0;
int dropCount = 0;
int mavPacketsReceived = 0;
uint8_t txPacketsSeqCounter = 0;

void udpOnPacketHandler(AsyncUDPPacket packet);

void WiFiEvent(WiFiEvent_t event);

MainController * MainController::_instance = NULL;
MainController & MainController::getInstance()
{
	if(_instance == NULL)
	{
		_instance = new MainController();
	}
	return *_instance;
}

MainController::MainController() : settingsManager(SettingsManager::getInstance())
{    
  for(int i = 0; i < MAX_TRACKED_TELEMETRY_COUNT; ++i)
  {
    _is_telemetry_enabled[i] = false;
    _is_telemetry_packet_received[i] = false;
  }
  for(int i = 0; i < MAX_SUPPORTED_DISTANCE_SENSORS_COUNT; ++i)
  {
    _current_dist_sensor_val[i] = 0;
  }

  _current_roll_rate = 0;
  _current_pitch_rate = 0;
  _current_yaw_rate = 0;

  _async_udp_blocked = false;
  mainControllerPtr = NULL;
  prevPacketSeq = 0;
  dropCount = 0;
  mavPacketsReceived = 0;
  _is_control_station_connected = 0;
  txPacketsSeqCounter = 0;
  _is_cmd_from_interrupt = false;
  
  attitudeControlMode = FusionMode;
  
  memset(mavSendBuffer, 0, sizeof(mavSendBuffer));

  isSomeConnectedBefore = false;
  noConnectSeconds = 0;

  currentMillis = millis();
  prevMillis = 0;
  
  memset(pixReceiverBuffer, 0, sizeof(pixReceiverBuffer));

  _missionPreparingReset();
}

void MainController::_missionPreparingReset()
{
  for(int i = 0; i < MAX_CLAW_COUNT; ++i)
  {
    _claw_angle[i] = 0;
  }
  
  _last_long_range_read_ts = 0;
  _last_short_range_read_ts = 0;
  _is_range_sensor_initialized = false;

  isManualControlOverrided = false;
  pitchControl = 0;
  rollControl = 0;
  throttleControl = 0;
  yawControl = 0;
  motorsArmCommandReceived = false;

  _current_pos_x = 0;
  _current_pos_y = 0;
  _current_pos_z = 0;
  _current_vel_x = 0;
  _current_vel_y = 0;
  _current_vel_z = 0;
  currentRoll = 0;
  currentPitch = 0; 
  currentYaw = 0;
  _current_roll_rate = 0;
  _current_pitch_rate = 0;
  _current_yaw_rate = 0;
  rcRoll = 1500;
  rcPitch = 1500;
  rcThrottle = 1500;
  rcYaw = 1500;

  wifiChannel = 0;

  _is_cmd_from_interrupt = false;

  _last_set_front_color_ms = 0;
  _last_set_back_color_ms = 0;
  _last_set_yaw_ms = 0;
  _yaw_relative_accumulator = 0;
  _last_yaw_accumulation_ms = 0;
  _last_send_log_ms = 0;
  _last_set_altitude_ms = 0;
  _altitude_relative_accumulator = 0;
  _last_altitude_accumulation_ms = 0;
  _last_set_target_pos_ms = 0;
  _pos_offset_accumulator[0] = _pos_offset_accumulator[1] = _pos_offset_accumulator[2] = 0;
  _last_pos_accumulation_ms = 0;
  _last_arming_ms = 0;
  _last_disarming_ms = 0;
  _last_takeoff_cmd_ms = 0;
  _last_land_cmd_ms = 0;
  _last_reboot_cmd_ms = 0;
  _last_headfree_cmd_ms = 0;

  for(int i = 0; i < MAX_SUPPORTED_SERVO_ID; ++i)
  {
    _last_set_servo_ms[i] = 0;
    _is_set_servo_angle_stuck[i] = false;
    _is_set_servo_pwm_stuck[i] = false;
  }
  for(int i = 0; i < MAX_SUPPORTED_MOTORS; ++i)
  {
    _last_set_motor_ms[i] = 0;
    _is_set_motor_stuck[i] = false;
  }

  _is_front_led_cmd_stuck = false;
  _is_back_led_cmd_stuck = false;
  _is_set_absolute_yaw_stuck = _is_set_relative_yaw_stuck = false;
  _is_relative_altitude_stuck = _is_absolute_altitude_stuck = false;
  _is_set_target_pos_stuck = _is_offset_pos_stuck = false;
  _is_arming_stuck = _is_disarming_stuck = false;
  _is_takeoff_cmd_stuck = _is_land_cmd_stuck = false;
  _is_reboot_cmd_stuck = false;
  _is_headfree_cmd_stuck = false;

  _yaw_cmd_ack_received = false;
  _altitude_cmd_ack_received = false;
  _position_cmd_ack_received = false;
  _arm_cmd_ack_received = false;
  _is_motionless = false;
}

MainController::~MainController()
{
  sendLogMsg(LIBRARY_SIGNATURE); // library signature for protect uploading
	if(_instance != NULL)
	{
		delete _instance;
	}
}

void MainController::_openAsyncUDP()
{
  if(Udp.listen(localUdpPort))
  {
/*#if DEBUG  
      Serial.println(WiFi.localIP());
#endif*/
      Udp.onPacket(udpOnPacketHandler);  
  }
}

void MainController::_closeAsyncUDP()
{
  Udp.close();
}

void startRetransmitTimer()
{
	os_timer_arm(&stmToAppRetransmitTimer, 1000 / RETRANSMIT_TIMER_FREQ, true);
}

void stopRetransmitTimer()
{
	os_timer_disarm(&stmToAppRetransmitTimer);
}

void retransmitTimerCallback(void *pArg) /// RETRANSMIT_TIMER_FREQ
{
   stopRetransmitTimer();
   MainController & main_controller = MainController::getInstance();
   if(main_controller._async_udp_blocked)
   {
      return;
   }
   main_controller._grabSerialAndRetransmit();
   
   if(main_controller._next_flush_stuck_commands_ms < millis())
   {
     main_controller._is_cmd_from_interrupt = true;
     main_controller._tryFlushStuckCommands();
     main_controller._is_cmd_from_interrupt = false;
     main_controller._next_flush_stuck_commands_ms += 1000 / FLUSH_STUCK_COMMANDS_HZ;
   }
   startRetransmitTimer();
}

void MainController::initializeSystem()
{
  settingsManager = SettingsManager::getInstance();

  uint8_t offer_router_mode = 0;

  mainControllerPtr = this;
  
  bool res;
  int ret = NANOPIX_STATUS_OK;
  IPAddress localIP;

  isSomeConnectedBefore = false;
  noConnectSeconds = 0;

  _battery_voltage = 0;
 
  //Serial1.begin(38400);
  //readerSerial.begin(38400);
#if NANOPIX_BOARD_V8_1
  STM_BootManager::reset();
#endif

  delay(500);

  srand(ESP.getCycleCount());
  
  FirmwareDataReceiver::getInstance().registerSystemUpdater(1, &EspUpdater::getInstance());
  FirmwareDataReceiver::getInstance().registerSystemUpdater(2, &StmUpdater::getInstance());
  FirmwareDataReceiver::getInstance().finishInitialization();

  if(!_esp_now_mode)
  {
    NP_CHECK(udpSender.initialize(&Udp));

    delay(200);
    WiFi.mode(WIFI_AP);
    WiFi.softAPConfig(udpSender.ip, udpSender.gateway, udpSender.subnet);

    wifiChannel = settingsManager.getWifiChannel();
    if(wifiChannel == 0)
      wifiChannel = selectOptimalWifiChannel();

    wifi_softap_set_dhcps_offer_option(OFFER_ROUTER, &offer_router_mode);
    res = WiFi.softAP(settingsManager.SSID.c_str(), settingsManager.PASS.c_str(), wifiChannel);
    //_debug_serial->printf("SoftAP %s:%s on channel %d", settingsManager.SSID.c_str(), settingsManager.PASS.c_str(), wifiChannel);
    //_debug_serial->println(res ? " READY" : " START FAILURE");

    //WiFi.softAP("quadcopter", "testpass");
    //WiFi.setOutputPower(10);
    
    _openAsyncUDP();
    
    STM_OTA_UDP_Receiver::start(STM32_OTA_UDP_PORT, STM32_OTA_SERVER_UDP_PORT);

    //WiFi.onEvent(WiFiEvent);
    localIP = WiFi.softAPIP();
  /*#if DEBUG  
    Serial.print("My IP: ");
    Serial.println(localIP.toString().c_str());
    Serial.print("Local IP: ");
    Serial.println(WiFi.localIP());
  #endif*/

    ArduinoOTA.onStart([]() {
      //Serial.println("Start");
      STM_BootManager::toResetState();
      
    });
    ArduinoOTA.onEnd([]() {
      //Serial.println("\nEnd");
      STM_BootManager::reset();
    });
    ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
      //Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
    });
    ArduinoOTA.onError([](ota_error_t error) {
      /*Serial.printf("Error[%u]: ", error);
      if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
      else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
      else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
      else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
      else if (error == OTA_END_ERROR) Serial.println("End Failed");*/
    });
    ArduinoOTA.begin();
  }
  os_timer_setfn(&stmToAppRetransmitTimer, retransmitTimerCallback, NULL);
  _next_flush_stuck_commands_ms = millis();
  startRetransmitTimer();
out:
    //if(ret != NANOPIX_STATUS_OK)
        //_debug_serial->printf("Initialization failure, code: %d", ret);
  delay(50);
  //Serial.end();
  Serial.begin(UART_BAUDRATE);
}

void MainController::_retransmitFromWireless(uint8_t * data, uint8_t length)
{
  uint16_t packet_size = 0;
  bool modified = false;
  bool is_esp_recipient = false;
  mavlink_message_t * resp_msg = NULL;
  int res_error_code = 0;
  if(mainControllerPtr != NULL)
  {
    for(int i = 0; i < length; i++) {
      uint8_t res = mavlink_parse_char(MAV_RX_CHAN, data[i], 
        &(mainControllerPtr->_rx_mav_msg), &(mainControllerPtr->_mav_status) );
      if(res) {
         mainControllerPtr->isSomeConnectedBefore = true; 
        
         uint8_t packID = mainControllerPtr->_rx_mav_msg.msgid;
         uint8_t packetSeq = mainControllerPtr->_rx_mav_msg.seq;
  
         _is_control_station_connected = true;
  
         mavPacketsReceived++;
         uint8_t dif = packetSeq - prevPacketSeq;
         dropCount += dif - 1;
  /* #if DEBUG      
         Serial.print(packID);
         Serial.print(" | ");
         Serial.print(mainControllerPtr->_rx_mav_msg.seq);
         Serial.print(" | ");
         Serial.write(packetData, dataLength);
         Serial.println();
  #endif*/
         
         prevPacketSeq = packetSeq;

         //DEBUG_PRINTF("mav mgs from wireless, id=%d\n", (int)packID);
         
         switch(packID) {
          case MAVLINK_MSG_ID_MANUAL_CONTROL:
            mavlink_manual_control_t manual_ctrl_msg;
            mavlink_msg_manual_control_decode(&(mainControllerPtr->_rx_mav_msg), &(manual_ctrl_msg) );

            mainControllerPtr->rcRoll = manual_ctrl_msg.x;
            mainControllerPtr->rcPitch = manual_ctrl_msg.y; 
            mainControllerPtr->rcThrottle = manual_ctrl_msg.z;
            mainControllerPtr->rcYaw = manual_ctrl_msg.r;
  
            // override data in manual control packet:
            //if(mainControllerPtr->isManualControlOverrided) {
            mavlink_message_t msg;
  
            int rollRC, pitchRC, throttleRC, yawRC;
            switch(mainControllerPtr->attitudeControlMode) {
              case MainController::OverrideMode:
                rollRC = mainControllerPtr->rollControl * 11 + 1500;
                pitchRC = mainControllerPtr->pitchControl * 11 + 1500;
                throttleRC = mainControllerPtr->throttleControl * 500 + 1500;
                yawRC = mainControllerPtr->yawControl * 11 + 1500;
              break;
              case MainController::FusionMode:
                rollRC = mainControllerPtr->rcRoll + mainControllerPtr->rollControl * 11;
                pitchRC = mainControllerPtr->rcPitch + mainControllerPtr->pitchControl * 11;
                throttleRC = mainControllerPtr->rcThrottle + mainControllerPtr->throttleControl * 500;
                yawRC = mainControllerPtr->rcYaw + mainControllerPtr->yawControl * 11;
              break;
            }
            
            mavlink_msg_manual_control_pack(0, 0, &msg, 0, rollRC, pitchRC, throttleRC, yawRC, 0);
            packet_size = mavlink_msg_to_send_buffer(mainControllerPtr->udpRxBuffer, &msg);
            modified = true;
            //}
  
            /*if( abs(mainControllerPtr->rcRoll - 1500) > 30 || abs(mainControllerPtr->rcPitch - 1500) > 30
            || abs(mainControllerPtr->rcThrottle - 1500) > 30 || abs(mainControllerPtr->rcYaw - 1500) > 30 )
            {
              if(mainControllerPtr->isManualControlOverrided) {
                mainControllerPtr->flightScriptIsBlocked = true;
                mainControllerPtr->isManualControlOverrided = false;
              }
            }*/
          break;
          case MAVLINK_MSG_ID_COMMAND_LONG:
            mavlink_command_long_t commandLongMsg;
            mavlink_msg_command_long_decode(&(mainControllerPtr->_rx_mav_msg), &(commandLongMsg) );
  
            if(commandLongMsg.command == MAV_CMD_COMPONENT_ARM_DISARM) 
            {
              if(commandLongMsg.param1 > 0.5f) // then arm command 
              {
                mainControllerPtr->motorsArmCommandReceived = true;
              }
            }
            
            if(commandLongMsg.command == MAV_CMD_MISSION_START) 
            {
              mainControllerPtr->missionStartCmdReceived = true;
            }
            
            if(commandLongMsg.command == MAV_CMD_PREFLIGHT_REBOOT_SHUTDOWN) 
            {
              if(commandLongMsg.param2 == 1)
              {
                ESP.restart();
              }
              if(commandLongMsg.param2 == 3)
              {
                mainControllerPtr->settingsManager.setNextBootMode(SettingsManager::UPGRADE_BOOT_MODE);
                ESP.restart();
              }
            }
            
          break;
#if NANOPIX_BOARD_V8_1 /// external STM reset for v8+ board:
         case MAVLINK_MSG_ID_COMMAND_INT:
            mavlink_command_int_t comInt;
            mavlink_msg_command_int_decode(&(mainControllerPtr->_rx_mav_msg), &comInt);

            if(comInt.command == MAV_CMD_PREFLIGHT_REBOOT_SHUTDOWN)
            {
              if(comInt.param1 == 1)
              {
                STM_BootManager::reset();
              }
              if(comInt.param1 == 2)
              {
                //EEPROM_Manager::eraseAll(); // TODO stm eeprom reset
                ///STM_BootManager::reset();
              }
            }
            break;
#endif
          case MAVLINK_MSG_ID_LOG_DATA:
            is_esp_recipient = true;
            resp_msg = NULL;
            /*if(mainControllerPtr->_rx_mav_msg.compid == 0)
            {
              res_error_code = _esp_ota_manager.handleEspOtaPacket(&(mainControllerPtr->_rx_mav_msg), &resp_msg);
            }*/
            /*if(mainControllerPtr->_rx_mav_msg.compid == 1)
            {
              res_error_code = _stm_mav_ota_manager.handleOtaPacket(&(mainControllerPtr->_rx_mav_msg), &resp_msg);
            }*/
            if(mainControllerPtr->_rx_mav_msg.compid == 3)
            {
              FirmwareDataReceiver::getInstance().handlePacket(&(mainControllerPtr->_rx_mav_msg), &resp_msg);
            }
            /*if((res_error_code != 0) || 
               (StmUpdater::getInstance().getLastHandleResult() == StmUpdater::HandleResult::FATAL_ERROR))
            {
              STM_BootManager::reset();
            }*/
            
            if(resp_msg != NULL)
            {
              int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, resp_msg);
              if(_esp_now_mode)
              {
                EspNowSlave::getInstance()->addDataToSend(mavSendBuffer, txSize);
              }
              else
              {
                udpSender.sendToConnected(mavSendBuffer, txSize);
              }
            }
            break;
          case MAVLINK_MSG_ID_REQUEST_DATA_STREAM:
            mavlink_request_data_stream_t requested_stream;
            mavlink_msg_request_data_stream_decode(&(mainControllerPtr->_rx_mav_msg), &requested_stream);
            if(requested_stream.req_stream_id > 0 && 
               requested_stream.req_stream_id <= MAX_TRACKED_TELEMETRY_COUNT)
            {
              _is_telemetry_enabled[requested_stream.req_stream_id - 1] = 
                                    requested_stream.start_stop && requested_stream.req_message_rate > 0;
            }
          break;
          default:
          
          break;
         }
         //if(!(packID == MAVLINK_MSG_ID_MANUAL_CONTROL/* && !mainControllerPtr->udpSender.isMainController(addr)*/))
         if(!is_esp_recipient)
         {
            if(modified)
            {
                Serial.write(mainControllerPtr->udpRxBuffer, packet_size);
            }
            else
            {
                Serial.write(data, length);
            }
         }
      }
    } 
  }
}

void udpOnPacketHandler(AsyncUDPPacket packet) {
  MainController & main_ctrl = MainController::getInstance();
  if(main_ctrl._async_udp_blocked)
  {
    return;
  }
    
  uint8_t* packetData = packet.data();
  size_t dataLength = packet.length();

  memcpy(main_ctrl.udpRxBuffer, packetData, dataLength);

  IPAddress addr = packet.remoteIP();

  if((mainControllerPtr != NULL) && dataLength)
  {
    main_ctrl._retransmitFromWireless(main_ctrl.udpRxBuffer, dataLength);
    if(!main_ctrl.udpSender.hasController(addr))
    {
        main_ctrl.udpSender.registerController(addr);
    }
  }
}

void MainController::_tryFlushStuckCommands()
{
  if(_is_front_led_cmd_stuck)
  {
    setFrontLEDColor(_stuck_front_led_color);
    if(!_is_front_led_cmd_stuck)
    {
      return;
    }
  }
  if(_is_back_led_cmd_stuck)
  {
    setBackLEDColor(_stuck_back_led_color);
    if(!_is_back_led_cmd_stuck)
    {
      return;
    }
  }

  for(int i = 0; i < MAX_SUPPORTED_MOTORS; ++i)
  {
    if(_is_set_motor_stuck[i])
    {
      setMotorSpeed(i + 1, _stuck_set_motor_value[i]);
      if(!_is_set_motor_stuck[i])
      {
        return;
      }
    }
  }
  
  if(_is_set_absolute_yaw_stuck)
  {
    setYawDegrees(_stuck_yaw_value[0], _stuck_yaw_value[1], false);
    if(!_is_set_absolute_yaw_stuck)
    {
      return;
    }
  }

  if(_is_set_relative_yaw_stuck)
  {
    rotateYawDegrees(0, _stuck_yaw_value[1], false);
    if(!_is_set_relative_yaw_stuck)
    {
      return;
    }
  }

  if(_is_absolute_altitude_stuck)
  {
    setTargetAltitude(_stuck_altitude_value[0], _stuck_altitude_value[1], false);
    if(!_is_absolute_altitude_stuck)
    {
      return;
    }
  }

  if(_is_relative_altitude_stuck)
  {
    offsetTargetAltitude(0, _stuck_altitude_value[1], false);
    if(!_is_relative_altitude_stuck)
    {
      return;
    }
  }

  for(int i = 0; i < MAX_SUPPORTED_SERVO_ID; ++i)
  {
    if(_is_set_servo_pwm_stuck[i])
    {
      setServoPWM(i + 1, _stuck_servo_value[i]);
      if(!_is_set_servo_pwm_stuck[i])
      {
        return;
      }
    }
  }

  for(int i = 0; i < MAX_SUPPORTED_SERVO_ID; ++i)
  {
    if(_is_set_servo_angle_stuck[i])
    {
      setServoAngle(i + 1, _stuck_servo_value[i]);
      if(!_is_set_servo_angle_stuck[i])
      {
        return;
      }
    }
  }

  if(_is_set_target_pos_stuck)
  {
    setTargetPos(_stuck_target_pos[0], _stuck_target_pos[1], _stuck_target_pos[2],
                 _stuck_target_pos[3], _stuck_target_pos[4], _stuck_target_pos[5], false);
    if(!_is_set_target_pos_stuck)
    {
      return;
    }
  }

  if(_is_offset_pos_stuck)
  {
    offsetTargetPos(0, 0, 0, _stuck_target_pos[3], _stuck_target_pos[4], _stuck_target_pos[5], false);
    if(!_is_offset_pos_stuck)
    {
      return;
    }
  }

  if(_is_arming_stuck)
  {
    armMotors(false);
    if(!_is_arming_stuck)
    {
      return;
    }
  }
  if(_is_disarming_stuck)
  {
    disarmMotors();
    if(!_is_disarming_stuck)
    {
      return;
    }
  }

  if(_is_takeoff_cmd_stuck)
  {
    takeoff(false);
    if(!_is_takeoff_cmd_stuck)
    {
      return;
    }
  }
  if(_is_land_cmd_stuck)
  {
    land(false);
    if(_is_land_cmd_stuck)
    {
      return;
    }
  }
  
  if(_is_reboot_cmd_stuck)
  {
    rebootFlightController();
    if(!_is_reboot_cmd_stuck)
    {
      return;
    }
  }

  if(_is_headfree_cmd_stuck)
  {
    _setHeadfree(_stuck_headfree_value);
    if(!_is_headfree_cmd_stuck)
    {
      return;
    }
  }
}

bool MainController::isMissionTriggered()
{
	bool val = missionStartCmdReceived;
	missionStartCmdReceived = false;
  if(val) 
  {
    _missionPreparingReset();
    _stored_headfree_state = isHeadfreeEnabled();
    disableHeadfree();
    _is_mission_running = true;
  }
	return val;
}

bool MainController::resetMissionTrigger()
{
  missionStartCmdReceived = false;
  _setHeadfree(_stored_headfree_state);
  motorsArmCommandReceived = false;
  _is_mission_running = false;
  disableAllServo();
}

void MainController::setAttitudeControlMode(AttitudeControlMode attitudeCtrlMode)
{
  attitudeControlMode = attitudeCtrlMode;
}

bool MainController::isArmCommandReceived()
{
  bool res = motorsArmCommandReceived;
  motorsArmCommandReceived = false;
  return res;
}

void MainController::setPitch(float deg_angle, bool pitch_only)
{
  isManualControlOverrided = true;
  pitchControl = constrain(deg_angle, -MAX_PITCH_ROLL_SETPOINT_DEG, MAX_PITCH_ROLL_SETPOINT_DEG);
  if(pitch_only) 
  {
    rollControl = 0;
    yawControl = 0;
    throttleControl = 0;
  }
}

void MainController::setRoll(float deg_angle, bool roll_only)
{
  isManualControlOverrided = true;
  rollControl = constrain(deg_angle, -MAX_PITCH_ROLL_SETPOINT_DEG, MAX_PITCH_ROLL_SETPOINT_DEG);
  if(roll_only)
  {
    pitchControl = 0;
    yawControl = 0;
    throttleControl = 0;
  }
}
void MainController::setYawSpeed(float speed_deg_sec)
{
  isManualControlOverrided = true;
  yawControl = constrain(speed_deg_sec, -MAX_YAW_RATE, MAX_YAW_RATE);
}

void MainController::_setYaw(float yaw_deg, float speed_deg_sec, bool is_relative, bool is_blocking)
{
  if(_is_cmd_from_interrupt) is_blocking = false;

  if(millis() < _last_set_yaw_ms + (1000 / MAX_SET_YAW_FREQ_HZ))
  {
    _stuck_yaw_value[1] = speed_deg_sec;
    if(is_relative)
    {
      _yaw_relative_accumulator += yaw_deg;
      _last_yaw_accumulation_ms = millis(); 
      _is_set_absolute_yaw_stuck = false;
      _is_set_relative_yaw_stuck = true;
    } 
    else 
    {
      _yaw_relative_accumulator = 0;
      _last_yaw_accumulation_ms = 0;
      _stuck_yaw_value[0] = yaw_deg;
      _is_set_relative_yaw_stuck = false;
      _is_set_absolute_yaw_stuck = true;
    }
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; // reject too high set yaw frequency
  }
  _is_set_relative_yaw_stuck = _is_set_absolute_yaw_stuck = false;

  _last_set_yaw_ms = millis();
  if(is_relative)
  {
    if(millis() < _last_yaw_accumulation_ms + OFFSET_ACCUMULATION_TIMEOUT_MS)
    {
      yaw_deg += _yaw_relative_accumulator;
      _last_yaw_accumulation_ms = 0;
    }
  }
  _yaw_relative_accumulator = 0;
  float yaw_delta = constrain(yaw_deg, -MAX_YAW_SETPOINT_DELTA_DEG, MAX_YAW_SETPOINT_DELTA_DEG);
  _yaw_cmd_ack_received = false;
  _sendMavCommandInt(MAV_CMD_CONDITION_YAW, yaw_deg, speed_deg_sec, 1, is_relative ? 1.0f : 0.0f);
  
  if(is_blocking)
  {
    uint32_t blocking_timeout = millis() + ((is_relative ? 2 * abs(yaw_deg) : 180) / 
                              max(double(MIN_YAW_SPEED_RAD * RAD_TO_DEG), double(abs(speed_deg_sec)))) * SEC_TO_MS;
    while(!_yaw_cmd_ack_received && (millis() < blocking_timeout))
    {
      delay(5);
    }
    _yaw_cmd_ack_received = false;
  }
}

void MainController::rotateYawDegrees(float delta_yaw_deg, float speed_deg_sec, bool is_blocking)
{
  _setYaw(delta_yaw_deg, speed_deg_sec, true, is_blocking);
}

void MainController::setYawDegrees(float target_yaw_deg, float speed_deg_sec, bool is_blocking)
{
  _setYaw(target_yaw_deg, speed_deg_sec, false, is_blocking);
}

void MainController::setThrottle(float throttle)
{
  isManualControlOverrided = true; 
  throttleControl = constrain(throttle, MIN_THROTTLE_VALUE, MAX_THROTTLE_VALUE);
}

void MainController::setServoAngle(int servo_number, int angle_deg)
{
  servo_number = constrain(servo_number, MIN_SERVO_NUMBER, MAX_SUPPORTED_SERVO_ID);
  uint8_t servo_id = servo_number - 1;
  if(millis() < _last_set_servo_ms[servo_id] + (1000 / MAX_SET_SERVO_FREQ_HZ)) 
  {
    _stuck_servo_value[servo_id] = angle_deg;
    _is_set_servo_angle_stuck[servo_id] = true;
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; // reject too high set servo frequency
  }
  _is_set_servo_angle_stuck[servo_id] = false;
  _last_set_servo_ms[servo_id] = millis();
  
  int angle = constrain(angle_deg, MIN_SERVO_ANGLE, MAX_SERVO_ANGLE);
  int pwm_value_us = 1000 + 1000 * (angle / 180.0f);
  setServoPWM(servo_number, pwm_value_us);
}

void MainController::openClaw(int index)
{
  setClaw(100, index);
}

void MainController::closeClaw(int index)
{
  setClaw(0, index);
}

void MainController::setClaw(int percent, int index)
{
  if(index < 0 || index >= MAX_CLAW_COUNT)
  {
    return;
  }
  
  percent = constrain(percent, 0, 100);
  int claw_interval = MAX_CLAW_SERVO_ANGLE - MIN_CLAW_SERVO_ANGLE;
  int target_angle = MIN_CLAW_SERVO_ANGLE + (percent * claw_interval) / 100;
  target_angle = constrain(target_angle, MIN_CLAW_SERVO_ANGLE, MAX_CLAW_SERVO_ANGLE);
  _claw_angle[index] = constrain(_claw_angle[index], MIN_CLAW_SERVO_ANGLE, MAX_CLAW_SERVO_ANGLE);
  if(_claw_angle[index] != target_angle)
  {
    int dir = _claw_angle[index] < target_angle ? 1 : -1;
    do
    {
      _claw_angle[index] += dir;
      setServoAngle(_claw_servo_ids[index], _claw_angle[index]);
      delay(10);
    } while (_claw_angle[index] != target_angle);
  }
}

void MainController::setServoPWM(int servo_number, int pwm_value_us)
{
  servo_number = constrain(servo_number, MIN_SERVO_NUMBER, MAX_SUPPORTED_SERVO_ID);
  uint8_t servo_id = servo_number - 1;
  if(millis() < _last_set_servo_ms[servo_id] + (1000 / MAX_SET_SERVO_FREQ_HZ)) 
  {
    _stuck_servo_value[servo_id] = pwm_value_us;
    _is_set_servo_pwm_stuck[servo_id] = true;
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; // reject too high set servo frequency
  }
  _is_set_servo_pwm_stuck[servo_id] = false;
  _last_set_servo_ms[servo_id] = millis();

  pwm_value_us = constrain(pwm_value_us, MIN_SERVO_PWM_US, MAX_SERVO_PWM_US);
  _sendMavCommandInt(MAV_CMD_DO_SET_SERVO, servo_number, pwm_value_us);
}

void MainController::disableServo(int servo_number)
{
  servo_number = constrain(servo_number, MIN_SERVO_NUMBER, MAX_SUPPORTED_SERVO_ID);
  _sendMavCommandInt(MAV_CMD_DO_SET_SERVO, servo_number, -2);
}

void MainController::disableAllServo()
{
  _sendMavCommandInt(MAV_CMD_DO_SET_SERVO, 0, -2);
}

void MainController::_setParam(const char* name, float value)
{
  mavlink_message_t msg;
  mavlink_msg_param_set_pack(0, 0, &msg, 0, 0, name, value, MAV_PARAM_TYPE_REAL32);
  int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &msg);
  noInterrupts();
  Serial.write(mavSendBuffer, txSize);
  interrupts();
}

int MainController::_sendMavlinkPacket(mavlink_message_t * msg, bool multicast)
{
    int ret = NANOPIX_STATUS_OK;
    int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, msg);
    if(_esp_now_mode)
    {
      EspNowSlave::getInstance()->addDataToSend(mavSendBuffer, txSize);
    }
    else
    {
      if(multicast || (udpSender.connectedCount() == 0))
      {
          NP_CHECK(udpSender.sendMulticast(mavSendBuffer, txSize));
      }
      else
      {
          NP_CHECK(udpSender.sendToConnected(mavSendBuffer, txSize));
      }
    }
out:
    return ret;
}

int MainController::sendLogMsg(const char* str, bool multicast)
{
  if(millis() < _last_send_log_ms + (1000 / MAX_SEND_LOG_FREQ_HZ)) // TODO: put logs to queue and send with intervals
  {
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return NANOPIX_STATUS_FULL; // reject too high log frequency
  }
  _last_send_log_ms = millis();
  int ret = NANOPIX_STATUS_OK;
  uint8_t chan = 1;
  static mavlink_message_t msg;
  mavlink_msg_statustext_pack_chan(0, 0, chan, &msg, 0, str);
  NP_CHECK(_sendMavlinkPacket(&msg, multicast));
out:
  return ret;
}

int MainController::sendLogMsg(int value, bool multicast)
{
  int ret = NANOPIX_STATUS_OK;
  char logbuf[32];
  sprintf(logbuf, "%d", value);
  NP_CHECK(sendLogMsg(logbuf, multicast));
out:
  return ret;
}

int MainController::sendLogMsg(float value, bool multicast)
{
  int ret = NANOPIX_STATUS_OK;
  char logbuf[32];

  float tmpVal = (value < 0) ? -value : value;

  int tmpInt1 = tmpVal;                  
  float tmpFrac = tmpVal - tmpInt1;   
  int tmpInt2 = trunc(tmpFrac * 10000);
  if(value < 0) 
  {
    sprintf(logbuf, "-%d.%04d", tmpInt1, tmpInt2);
  }
  else
  {
    sprintf(logbuf, "%d.%04d", tmpInt1, tmpInt2);
  }

  NP_CHECK(sendLogMsg(logbuf, multicast));
out:
  return ret;
}

void MainController::sendToUART(const uint8_t* data, uint16_t count)
{
  static mavlink_message_t msg;
  int index = 0;
  while(index < count)
  {
    int datasize = MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN;
    if(count - index < MAVLINK_MSG_LOG_DATA_FIELD_DATA_LEN)
      datasize = count - index;
    mavlink_msg_log_data_pack_chan(0, 0, 0, &msg, 0, 0, datasize, &(data[index]));
    int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &msg);
    noInterrupts();
    Serial.write(mavSendBuffer, txSize);
    interrupts();
    index += datasize;
  }

}

void MainController::setLEDsColors(uint8_t forwardR, uint8_t forwardG, uint8_t forwardB, uint8_t backwardR, uint8_t backwardG, uint8_t backwardB)
{
  setFrontLEDColor(forwardR, forwardG, forwardB);
  setBackLEDColor(backwardR, backwardG, backwardB);
}

void MainController::setLEDsColors(Color & forwardColor, Color & backwardColor)
{
  setLEDsColors(forwardColor.R, forwardColor.G, forwardColor.B, backwardColor.R, backwardColor.G, backwardColor.B);
}

void MainController::setFrontLEDColor(uint8_t forwardR, uint8_t forwardG, uint8_t forwardB)
{
  if(millis() < _last_set_front_color_ms + (1000 / MAX_SET_COLORS_FREQ_HZ))
  {
    _stuck_front_led_color = Color(forwardR, forwardG, forwardB);
    _is_front_led_cmd_stuck = true;
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; // reject too high set color frequency
  }
  _is_front_led_cmd_stuck = false;
  _last_set_front_color_ms = millis();
  LedColor fc;
  fc.R = forwardR;
  fc.G = forwardG;
  fc.B = forwardB;
  fc.mode = COLOR_NOT_SAVE_MODE;
  _setParam("FrontLed", fc.floatParamValue);
}

void MainController::setBackLEDColor(uint8_t backwardR, uint8_t backwardG, uint8_t backwardB)
{
  if(millis() < _last_set_back_color_ms + (1000 / MAX_SET_COLORS_FREQ_HZ))
  {
    _stuck_back_led_color = Color(backwardR, backwardG, backwardB);
    _is_back_led_cmd_stuck = true;
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; // reject too high set color frequency
  }
  _is_back_led_cmd_stuck = false;
  _last_set_back_color_ms = millis();
  LedColor bc;
  bc.R = backwardR;
  bc.G = backwardG;
  bc.B = backwardB;
  bc.mode = COLOR_NOT_SAVE_MODE;
  _setParam("BackLed", bc.floatParamValue);
}

void MainController::setFrontLEDColor(Color & forwardColor)
{
	setFrontLEDColor(forwardColor.R, forwardColor.G, forwardColor.B);
}

void MainController::setBackLEDColor(Color & backwardColor)
{
  setBackLEDColor(backwardColor.R, backwardColor.G, backwardColor.B);
}

void MainController::_switchArmDisarm(bool armState, bool is_blocking)
{
  if(_is_cmd_from_interrupt) is_blocking = false;

  if(armState) 
  {
    if(millis() < _last_arming_ms + (1000 / MAX_COMMANDS_FREQ_HZ))
    {
      _is_disarming_stuck = false;
      _is_arming_stuck = true;
      if(!_is_cmd_from_interrupt)
      {
        delay(1);
      }
      return; 
    }
    _last_arming_ms = millis();
  } 
  else 
  {
    if(millis() < _last_disarming_ms + (1000 / MAX_COMMANDS_FREQ_HZ))
    {
      _is_arming_stuck = false;
      _is_disarming_stuck = true;
      if(!_is_cmd_from_interrupt)
      {
        delay(1);
      }
      return;
    }
    _last_disarming_ms = millis();
  }
  _is_disarming_stuck = _is_arming_stuck = false;
  
  mavlink_msg_command_long_pack(0, 0, &_tx_mav_msg, 0, 0, MAV_CMD_COMPONENT_ARM_DISARM, 
                                0, (armState ? 1.0f : 0.0f), 0, 0, 0, 0, 0, 0);
  int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &_tx_mav_msg);
  
  noInterrupts();
  _arm_cmd_ack_received = false;
  Serial.write(mavSendBuffer, txSize);
  interrupts();

  if(is_blocking)
  {
    uint32_t blocking_timeout = millis() + ARM_CMD_TIMEOUT_SEC * SEC_TO_MS;
    while(!_arm_cmd_ack_received && (millis() < blocking_timeout))
    {
      delay(5);
    }
    _arm_cmd_ack_received = false;
  }
}

void MainController::setMotorSpeed(uint8_t motor_number, float speed_percent)
{
  if((motor_number == 0) || (motor_number > MAX_SUPPORTED_MOTORS))
  {
    return;
  }
  uint8_t mot_id = motor_number - 1;

  if(millis() < _last_set_motor_ms[mot_id] + (1000 / MAX_SET_MOTORS_FREQ_HZ))
  {
    _stuck_set_motor_value[mot_id] = speed_percent;
    _is_set_motor_stuck[mot_id] = true;
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; // reject too high set motors frequency
  }
  _is_set_motor_stuck[mot_id] = false;
  _last_set_motor_ms[mot_id] = millis();
  mavlink_msg_command_long_pack(0, 0, &_tx_mav_msg, 0, 0, MAV_CMD_DO_MOTOR_TEST, 
                                0, motor_number, MOTOR_TEST_THROTTLE_PERCENT, 
                                constrain(speed_percent, 0.0f, 100.0f), 0, 0, 0, 0);
  int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &_tx_mav_msg);
  noInterrupts();
  Serial.write(mavSendBuffer, txSize);
  interrupts();
}

void MainController::armMotors(bool is_blocking)
{
  _switchArmDisarm(true, is_blocking);
}

void MainController::disarmMotors()
{
  _switchArmDisarm(false, false);
}

void MainController::_sendMavCommandInt(uint8_t msgType, 
                                          float param1, 
                                          float param2, 
                                          float param3, 
                                          float param4, 
                                          float param5, 
                                          float param6, 
                                          float param7)
{
  mavlink_msg_command_int_pack(0, 0, &_tx_mav_msg, 0, 0, 0, msgType, 0, 0, 
                               param1, param2, param3, param4, param5, param6, param7);
  int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &_tx_mav_msg);
  noInterrupts();
  Serial.write(mavSendBuffer, txSize);
  interrupts();
}

void MainController::rebootFlightController()
{
  if(millis() < _last_reboot_cmd_ms + (1000 / MAX_COMMANDS_FREQ_HZ))
  {
    _is_reboot_cmd_stuck = true;
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; 
  }
  _is_reboot_cmd_stuck = false;
  _last_reboot_cmd_ms = millis();
  _sendMavCommandInt(MAV_CMD_PREFLIGHT_REBOOT_SHUTDOWN, 1);
}

void MainController::_setAltitude(float meters, float max_speed, bool is_relative, bool is_blocking)
{
  if(_is_cmd_from_interrupt) is_blocking = false;

  if(millis() < _last_set_altitude_ms + (1000 / MAX_SET_ALTITUDE_FREQ_HZ))
  {
    _stuck_altitude_value[1] = max_speed;
    if(is_relative)
    {
      _last_altitude_accumulation_ms = millis();
      _altitude_relative_accumulator += meters;
      _is_absolute_altitude_stuck = false;
      _is_relative_altitude_stuck = true;
    } 
    else
    {
      _stuck_altitude_value[0] = meters;
      _last_altitude_accumulation_ms = 0;
      _altitude_relative_accumulator = 0;
      _is_relative_altitude_stuck = false;
      _is_absolute_altitude_stuck = true;
    }
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; // reject too high set altitude frequency
  }
  _is_relative_altitude_stuck = _is_absolute_altitude_stuck = false;

  _last_set_altitude_ms = millis();
  if(is_relative)
  {
    if(millis() < _last_altitude_accumulation_ms + OFFSET_ACCUMULATION_TIMEOUT_MS)
    {
      meters += _altitude_relative_accumulator;
      _last_altitude_accumulation_ms = 0;
    }
  }
  _altitude_relative_accumulator = 0;

  _altitude_cmd_ack_received = false;
  _sendMavCommandInt(MAV_CMD_DO_CHANGE_ALTITUDE, meters, 
                     MAV_FRAME_LOCAL_NED, is_relative ? 1.0f : 0.0f, max_speed);

  if(is_blocking)
  {
    uint32_t blocking_timeout = millis() + 
      (is_relative ? (5 * abs(meters)) / MIN_ALTITUDE_SPEED : ALTITUDE_CMD_TIMEOUT_SEC) * SEC_TO_MS;
    while(!_altitude_cmd_ack_received && (millis() < blocking_timeout))
    {
      delay(5);
    }
    _altitude_cmd_ack_received = false;
  }
}

void MainController::setTargetAltitude(float meters, float max_speed, bool is_blocking)
{
  meters = constrain(meters, 
                     MIN_ALTITUDE_SETPOINT_MM * MM_TO_METERS, 
                     MAX_ALTITUDE_SETPOINT_MM * MM_TO_METERS);
  _setAltitude(meters, max_speed, false, is_blocking);
}

void MainController::offsetTargetAltitude(float delta_meters, float max_speed, bool is_blocking)
{
  delta_meters = constrain(delta_meters, -MAX_ALTITUDE_OFFSET, MAX_ALTITUDE_OFFSET);
  _setAltitude(delta_meters, max_speed, true, is_blocking);
}

bool MainController::isHeadfreeEnabled()
{
  return _is_headfree_enabled;
}

void MainController::_setHeadfree(bool is_enabled)
{
  if(millis() < _last_headfree_cmd_ms + (1000 / MAX_COMMANDS_FREQ_HZ))
  {
    _stuck_headfree_value = is_enabled;
    _is_headfree_cmd_stuck = true;
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; 
  }
  _is_headfree_cmd_stuck = false;
  _last_headfree_cmd_ms = millis();
  
  _setParam("Headfree", is_enabled ? 1.0f : 0.0f);
}

void MainController::enableHeadfree()
{
  _setHeadfree(true);
}

void MainController::disableHeadfree()
{
  _setHeadfree(false);
}

bool MainController::isMotionless()
{
  return _is_motionless;
}

void MainController::_switchTakeoffLand(bool takeoff, bool is_blocking)
{
  if(_is_cmd_from_interrupt) is_blocking = false;

  mavlink_message_t msg;
  uint8_t commandByte;
  if(takeoff)
  {
      commandByte = MAV_CMD_NAV_TAKEOFF;
      if(millis() < _last_takeoff_cmd_ms + (1000 / MAX_COMMANDS_FREQ_HZ))
      {
        _is_land_cmd_stuck = false;
        _is_takeoff_cmd_stuck = true;
        if(!_is_cmd_from_interrupt)
        {
          delay(1);
        }
        return; 
      }
      _last_takeoff_cmd_ms = millis();
  }
  else
  {
      commandByte = MAV_CMD_NAV_LAND;
      if(millis() < _last_land_cmd_ms + (1000 / MAX_COMMANDS_FREQ_HZ))
      {
        _is_takeoff_cmd_stuck = false;
        _is_land_cmd_stuck = true;
        if(!_is_cmd_from_interrupt)
        {
          delay(1);
        }
        return;
      }
      _last_land_cmd_ms = millis();
  }
  _is_land_cmd_stuck = _is_takeoff_cmd_stuck = false;

  mavlink_msg_command_long_pack(0, 0, &msg, 0, 0, commandByte, 0, 0, 0, 0, 0, 0, 0, 0);
  int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &msg);
  noInterrupts();
  _takeoff_land_ack_received = false;
  Serial.write(mavSendBuffer, txSize);
  interrupts();

  if(is_blocking)
  {
    uint32_t blocking_timeout = millis() + TAKEOFF_CMD_TIMEOUT_SEC * SEC_TO_MS;
    while(!_takeoff_land_ack_received && (millis() < blocking_timeout))
    {
      delay(5);
    }
    _takeoff_land_ack_received = false;
  }
}

void MainController::requestTelemetryDataStream(TelemetryPacketType streamType, int streamRateHz, bool startStopFlag)
{
  mavlink_message_t msg;
  int stream_rate_hz = constrain(streamRateHz, 0, MAX_TELEMETRY_RATE_HZ);
  mavlink_msg_request_data_stream_pack(0, 0, &msg, 0, 0, streamType, stream_rate_hz, startStopFlag);
  int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &msg);
  noInterrupts();
  Serial.write(mavSendBuffer, txSize);
  interrupts();
}

void MainController::takeoff(bool is_blocking)
{
  setPitch(0);
  setRoll(0);
  setYawSpeed(0);
  _switchTakeoffLand(true, is_blocking);
}

void MainController::land(bool is_blocking)
{
  setPitch(0);
  setRoll(0);
  setYawSpeed(0);
  _switchTakeoffLand(false, is_blocking);
}

void MainController::_setPosition(float x, float y, float z, 
                                  float vel_x, float vel_y, float vel_z,
                                  bool is_relative, bool is_blocking)
{
  if(_is_cmd_from_interrupt) is_blocking = false;

  vel_x = constrain(vel_x, -MAX_POS_XY_SPEED, MAX_POS_XY_SPEED);
  vel_y = constrain(vel_y, -MAX_POS_XY_SPEED, MAX_POS_XY_SPEED);
  vel_z = constrain(vel_z, -MAX_POS_Z_SPEED, MAX_POS_Z_SPEED);

  if(millis() < _last_set_target_pos_ms + (1000 / MAX_SET_POS_FREQ_HZ))
  {
    _stuck_target_pos[3] = vel_x;
    _stuck_target_pos[4] = vel_y;
    _stuck_target_pos[5] = vel_z;
    if(is_relative)
    {
       _pos_offset_accumulator[0] += x;
       _pos_offset_accumulator[1] += y;
       _pos_offset_accumulator[2] += z;
       _last_pos_accumulation_ms = millis();
       _is_set_target_pos_stuck = false;
       _is_offset_pos_stuck = true;
    }
    else
    {
      _pos_offset_accumulator[0] = 0;
      _pos_offset_accumulator[1] = 0;
      _pos_offset_accumulator[2] = 0;
      _last_pos_accumulation_ms = 0;
      _stuck_target_pos[0] = x;
      _stuck_target_pos[1] = y;
      _stuck_target_pos[2] = z;
      _is_offset_pos_stuck = false;
      _is_set_target_pos_stuck = true;
    }
    if(!_is_cmd_from_interrupt)
    {
      delay(1);
    }
    return; // reject too high set altitude frequency
  }
  _is_offset_pos_stuck = _is_set_target_pos_stuck = false;
  _last_set_target_pos_ms = millis();

  if(is_relative)
  {
    if(millis() < _last_pos_accumulation_ms + OFFSET_ACCUMULATION_TIMEOUT_MS)
    {
       x += _pos_offset_accumulator[0];
       y += _pos_offset_accumulator[1];
       z += _pos_offset_accumulator[2];
      _last_pos_accumulation_ms = 0;
    }
  }
  _pos_offset_accumulator[0] = 0;
  _pos_offset_accumulator[1] = 0;
  _pos_offset_accumulator[2] = 0;
  
  float yaw_rate = 0, yaw = 0;
  uint8_t frame = is_relative ? MAV_FRAME_LOCAL_OFFSET_NED : MAV_FRAME_LOCAL_NED;

  mavlink_msg_set_position_target_local_ned_pack(0, 0, &_tx_mav_msg, millis(), 0, 0, frame, 0, 
                                                 x, y, z, vel_x, vel_y, vel_z, 
                                                 0, 0, 0, yaw, yaw_rate);
  int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &_tx_mav_msg);
  noInterrupts();
  _position_cmd_ack_received = false;
  Serial.write(mavSendBuffer, txSize);
  interrupts();

  bool is_altitude_set = fabs(vel_z) > 0.0001f;

  if(is_blocking)
  {
    _sendMavCommandInt(MAV_CMD_CONDITION_DISTANCE, POS_CMD_TOLERANCE, is_altitude_set ? 1.0f : 0.5f);

    uint32_t blocking_timeout = millis() + POSITION_CMD_TIMEOUT_SEC * SEC_TO_MS; // TODO more precise timeout
    while(!_position_cmd_ack_received && (millis() < blocking_timeout))
    {
      delay(5);
    }
    _position_cmd_ack_received = false;
  }
}

void MainController::setTargetPos(float pos_x, float pos_y, float pos_z, 
                    float vel_x, float vel_y, float vel_z,
                    bool is_blocking)
{
  _setPosition(pos_x, pos_y, pos_z, vel_x, vel_y, vel_z, false, is_blocking);
}

void MainController::offsetTargetPos(float pos_dx, float pos_dy, float pos_dz, 
                       float vel_x, float vel_y, float vel_z,
                       bool is_blocking)
{
  pos_dx = constrain(pos_dx, -MAX_POS_XY_OFFSET, MAX_POS_XY_OFFSET);
  pos_dy = constrain(pos_dy, -MAX_POS_XY_OFFSET, MAX_POS_XY_OFFSET);
  pos_dz = constrain(pos_dz, -MAX_ALTITUDE_OFFSET, MAX_ALTITUDE_OFFSET);
  _setPosition(pos_dx, pos_dy, pos_dz, vel_x, vel_y, vel_z, true, is_blocking);
}

void MainController::setTargetPos(float pos_x, float pos_y, bool is_blocking)
{
  setTargetPos(pos_x, pos_y, 0, DEFAULT_POS_SPEED, DEFAULT_POS_SPEED, 0, is_blocking);
}

void MainController::offsetTargetPos(float pos_dx, float pos_dy, bool is_blocking)
{
  offsetTargetPos(pos_dx, pos_dy, 0, DEFAULT_POS_SPEED, DEFAULT_POS_SPEED, 0, is_blocking);
}

void MainController::moveForward(float dist, float speed, bool is_blocking)
{
  offsetTargetPos(0, dist, 0, 0, speed, 0, is_blocking);
}

void MainController::moveBackward(float dist, float speed, bool is_blocking)
{
  offsetTargetPos(0, -dist, 0, 0, -speed, 0, is_blocking);
}

void MainController::moveLeft(float dist, float speed, bool is_blocking)
{
  offsetTargetPos(-dist, 0, 0, -speed, 0, 0, is_blocking);
}

void MainController::moveRight(float dist, float speed, bool is_blocking)
{
  offsetTargetPos(dist, 0, 0, speed, 0, 0, is_blocking);
}

void MainController::_grabSerialAndRetransmit()
{  
    int availableBytes = Serial.available();
    int pixReceiverCount = 0;
    if(availableBytes > 0)
    {
      pixReceiverCount = availableBytes > PIX_RECEIVER_BUFFER_SIZE ? PIX_RECEIVER_BUFFER_SIZE : availableBytes;
      Serial.readBytes(pixReceiverBuffer, pixReceiverCount);
    }

    for(int i = 0; i < pixReceiverCount; i++) 
    {
        uint8_t packetCompleted = mavlink_parse_char(MAV_TX_CHAN, pixReceiverBuffer[i], &(_uart_rx_mav_msg), &(_uart_mav_status) );

        if(packetCompleted) 
        {
            uint8_t packID = _uart_rx_mav_msg.msgid;
            switch(packID) {
            case MAVLINK_MSG_ID_LOCAL_POSITION_NED:
                mavlink_local_position_ned_t localPosMsg;
                mavlink_msg_local_position_ned_decode(&(_uart_rx_mav_msg), &localPosMsg);
                
                _current_pos_x = localPosMsg.x;
                _current_pos_y = localPosMsg.y;
                _current_pos_z = localPosMsg.z;

                _current_vel_x = localPosMsg.vx;
                _current_vel_y = localPosMsg.vy;
                _current_vel_z = localPosMsg.vz;

                _is_telemetry_packet_received[LocalPosition - 1] = true;
            break;
            case MAVLINK_MSG_ID_ATTITUDE:
                mavlink_attitude_t attitudeMsg;
                mavlink_msg_attitude_decode(&(_uart_rx_mav_msg), &attitudeMsg);
          
                currentRoll = attitudeMsg.roll;
                currentPitch = attitudeMsg.pitch;
                currentYaw = attitudeMsg.yaw;

                _current_roll_rate = attitudeMsg.rollspeed;
                _current_pitch_rate = attitudeMsg.pitchspeed;
                _current_yaw_rate = attitudeMsg.yawspeed;

                _is_telemetry_packet_received[AttitudePacket - 1] = true;
            break;
            case MAVLINK_MSG_ID_RAW_IMU:
                mavlink_raw_imu_t raw_imu_msg;
                mavlink_msg_raw_imu_decode(&(_uart_rx_mav_msg), &raw_imu_msg);
                
                _current_accel_x = raw_imu_msg.xacc;
                _current_accel_y = raw_imu_msg.yacc;
                _current_accel_z = raw_imu_msg.zacc;

                _current_gyro_x = raw_imu_msg.xgyro;
                _current_gyro_y = raw_imu_msg.ygyro;
                _current_gyro_z = raw_imu_msg.zgyro;
                _is_telemetry_packet_received[Standart9AxisIMU - 1] = true;
            break;
            case MAVLINK_MSG_ID_DISTANCE_SENSOR:
                mavlink_distance_sensor_t dist_sensor;
                mavlink_msg_distance_sensor_decode(&(_uart_rx_mav_msg), &dist_sensor);
                if(dist_sensor.id < MAX_SUPPORTED_DISTANCE_SENSORS_COUNT)
                {
                  _current_dist_sensor_val[dist_sensor.id] = dist_sensor.current_distance * MM_TO_METERS;
                  _is_telemetry_packet_received[DistanceSensor - 1] = true;
                }
            break;
            case MAVLINK_MSG_ID_LOG_DATA:
                mavlink_log_data_t logDataMsg;
                mavlink_msg_log_data_decode(&(_uart_rx_mav_msg), &logDataMsg);
                for(int i = 0; i < logDataMsg.count; i++)
                {
                    if(!_uart_bridge_buffer.add(logDataMsg.data[i]))
                    {
                      sendLogMsg("UART bridge RX buffer overflowed");
                    }
                }
            break;
            case MAVLINK_MSG_ID_SYS_STATUS:
                mavlink_sys_status_t sys_status_msg;
                mavlink_msg_sys_status_decode(&(_uart_rx_mav_msg), &sys_status_msg);
                _battery_voltage = sys_status_msg.voltage_battery / 1000.0f;
            break;
            case MAVLINK_MSG_ID_HEARTBEAT:
                mavlink_heartbeat_t heartbeat_msg;
                mavlink_msg_heartbeat_decode(&(_uart_rx_mav_msg), &heartbeat_msg);
                _is_headfree_enabled = heartbeat_msg.system_status & 0x80;
                _is_motionless = heartbeat_msg.system_status & 0x10;

                if(_is_mission_running)
                {
                  heartbeat_msg.custom_mode |= 0x10000000;
                }
                else
                {
                  heartbeat_msg.custom_mode &= 0xEFFFFFFF;
                }
                mavlink_msg_heartbeat_encode_chan(0, 0, 1, &_uart_rx_mav_msg, &heartbeat_msg);
            break;
            case MAVLINK_MSG_ID_COMMAND_ACK:
                mavlink_command_ack_t ack_msg;
                mavlink_msg_command_ack_decode(&(_uart_rx_mav_msg), &ack_msg);
                if(ack_msg.result == MAV_RESULT_ACCEPTED) 
                {
                  switch(ack_msg.command) {
                    case MAV_CMD_CONDITION_YAW:
                        _yaw_cmd_ack_received = true;
                    break;
                    case MAV_CMD_DO_CHANGE_ALTITUDE:
                        _altitude_cmd_ack_received = true;
                    break;
                    case MAV_CMD_CONDITION_DISTANCE:
                        _position_cmd_ack_received = true;
                    break;
                    case MAV_CMD_NAV_LAND:
                    case MAV_CMD_NAV_TAKEOFF:
                        _takeoff_land_ack_received = true;
                    break;
                    case MAV_CMD_COMPONENT_ARM_DISARM:
                        _arm_cmd_ack_received = true;
                    break;
                    default:
                    break;
                  }
                }
            break;
            default:
            break;
            }

            int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, &(_uart_rx_mav_msg) );
            if(!_esp_now_mode)
            {
              if (udpSender.connectedCount() == 0)
                  udpSender.sendMulticast(mavSendBuffer, txSize);
              else
                  udpSender.sendToConnected(mavSendBuffer, txSize);
            }
            else
            {
              EspNowSlave::getInstance()->addDataToSend(mavSendBuffer, txSize);
            }
        
        }
    }
    if(_esp_now_mode)
    {
      EspNowSlave::getInstance()->flush();
    }
}

bool MainController::_enableAndWaitTelemetry(TelemetryPacketType stream_type, uint16_t stream_rate, uint32_t timeout_ms)
{
    uint16_t stream_id = (int)stream_type - 1;
    if(!_is_telemetry_enabled[stream_id])
    {
      _is_telemetry_packet_received[stream_id] = false;
      requestTelemetryDataStream(stream_type, stream_rate, true);
      uint32_t timeout_ts = millis() + timeout_ms;
      while(!_is_telemetry_packet_received[stream_id] && (millis() < timeout_ts))
      {
        if(!_is_cmd_from_interrupt)
        {
          delay(1);
        }
      }
      if(_is_telemetry_packet_received[stream_id])
      {
        return true;
      } 
      else
      {
        return false;
      }
    }
    else
    {
      return true;
    }
}

size_t MainController::availableBytesOnUART()
{
    return _uart_bridge_buffer.numElements();
}

void MainController::initializeLineSensor(int digital_pin, bool with_power)
{
    pinMode(digital_pin, INPUT);
    if(with_power)
    {
      pinMode(4, OUTPUT);
      digitalWrite(4, HIGH);
      pinMode(12, OUTPUT);
      digitalWrite(12, LOW);
    }
}

bool MainController::readLineSensorValue(int digital_pin)
{
  return digitalRead(digital_pin);
}

float MainController::readExternalRangeSensorValue()
{
  if(!_is_range_sensor_initialized)
  {
    _range_sensor_type = UNDEF_TYPE;
    bool is_short_range = _initializeShortRangeSensor();
    if(is_short_range)
    {
      _range_sensor_type = VL53L0X_TYPE;
      _is_range_sensor_initialized = true;
    }
    else
    {
       bool is_long_range = _initializeLongRangeSensor();
       if(is_long_range)
       {
         _range_sensor_type = VL53L1X_TYPE;
         _is_range_sensor_initialized = true;
       }
    }
  }

  switch(_range_sensor_type)
  {
    case VL53L0X_TYPE:
      return _readShortRangeSensorValue();
    break;
    case VL53L1X_TYPE:
      return _readLongRangeSensorValue();
    break;
    default:
      return -1;
    break;
  }
}

bool MainController::_initializeShortRangeSensor()
{
  pinMode(2, OUTPUT);
  digitalWrite(2, HIGH);
  pinMode(15, OUTPUT);
  digitalWrite(15, LOW);
  Wire.begin(12, 4);
  Wire.setClock(400000); // use 400 kHz I2C

  _range_sensor.setAddress(0b0101001);
  _range_sensor.setTimeout(500);
  if (!_range_sensor.init())
  {
    sendLogMsg("Failed to detect and initialize VL53L0X range sensor!");
    return false;
  }
  _range_sensor.startContinuous(25);
  return true;
}

float MainController::_readShortRangeSensorValue()
{
  int32_t dt = millis() - _last_short_range_read_ts;
  if((dt >= 0) && (dt < 25))
  {
    delay(25 - dt);
  }
  _last_short_range_read_ts = millis();
  return _range_sensor.readRangeContinuousMillimeters() / 10.0f;
}

bool MainController::_initializeLongRangeSensor()
{
  pinMode(2, OUTPUT);
  digitalWrite(2, HIGH);
  pinMode(15, OUTPUT);
  digitalWrite(15, LOW);
  Wire.begin(12, 4);
  Wire.setClock(400000); // use 400 kHz I2C

  _long_range_sensor.setAddress(0b0101001);
  _long_range_sensor.setTimeout(500);
  if (!_long_range_sensor.init())
  {
    sendLogMsg("Failed to detect and initialize VL53L1X range sensor!");
    return false;
  }
  _long_range_sensor.setDistanceMode(VL53L1X::Long);
  _long_range_sensor.setMeasurementTimingBudget(50000);

  _long_range_sensor.startContinuous(25);
  return true;
}

float MainController::_readLongRangeSensorValue()
{
  int32_t dt = millis() - _last_long_range_read_ts;
  if((dt >= 0) && (dt < 25))
  {
    delay(25 - dt);
  }
  _last_long_range_read_ts = millis();
  return _long_range_sensor.read() / 10.0f;
}

uint16_t MainController::readFromUART(uint8_t * readBuffer, uint16_t readCount)
{
    uint16_t count = 0;
    for(int i = 0; i < readCount && !_uart_bridge_buffer.isEmpty(); i++)
    {
         _uart_bridge_buffer.pull(&readBuffer[i]);
         count = i + 1;
    }
    return count;
}

void MainController::_mainServiceRoutine()
{
  if(!_esp_now_mode)
  {
    if(STM_OTA_UDP_Receiver::isOTA_Request())
    {
      stopRetransmitTimer();
      _async_udp_blocked = true;
      STM_OTA_UDP_Receiver::handle();
      
      Serial.end();
      Serial.begin(UART_BAUDRATE);

      _async_udp_blocked = false;
      startRetransmitTimer();
      STM_OTA_UDP_Receiver::restart();
    }
    
    ArduinoOTA.handle();

    settingsManager.parseMavPackets();

    currentMillis = millis();
    if (currentMillis - prevMillis >= STATUS_REPORT_INTERVAL_MS) 
    {
      if(isSomeConnectedBefore == false) 
      {
        noConnectSeconds++;
        //settingsManager.sendString( "No connect seconds: " + String(noConnectSeconds) );
        //Serial.println(noConnectSeconds);
        if( (noConnectSeconds > RESET_SSID_PASS_SECONDS) && (!settingsManager.isDefaultPasswordNow()) )
        {
          settingsManager.resetSSID_PassToDefault();
          STM_BootManager::toResetState();
          delay(2000);
          STM_BootManager::reset();
          ESP.restart();
        }
      } else
        noConnectSeconds = 0;

      int dropRate = 0;
      if(mavPacketsReceived > 0)
        dropRate = 100 * (dropCount / (float)mavPacketsReceived);
            
      char dropLog[128];
      sprintf(dropLog, "ESP Rx: %d, drop rate: %d% : ch: %d", mavPacketsReceived, dropRate, wifiChannel);
      sendLogMsg(dropLog, true);
        
    //Serial1.println("serial1");
      
  /*#if DEBUG
      Serial.print("----------------- UDP PACKETS RECEIVED: ");
      Serial.print(mavPacketsReceived);
      Serial.print(" | DROPED: ");
      Serial.print(dropCount);
      Serial.print(" | DROP RATE: ");
      Serial.print(dropRate);
      Serial.println("%");
  #endif*/
      dropCount = 0;
      mavPacketsReceived = 0;

      prevMillis = millis();
    }
  }

  /*while(readerSerial.available())
  {
	Serial1.write(readerSerial.read());
  }*/

  /*int bytesToRead = availableBytesOnUART();
  if(bytesToRead > 0)
  {
      uint8_t readBuf[64];
      int readCount = readFromUART(readBuf, 64);
      //sendLogMsg((char*)readBuf);
      sendToUART(readBuf, readCount);
  }*/

  IUpdater::HandleResult esp_handle_res = EspUpdater::getInstance().handle();
  if(esp_handle_res == IUpdater::FATAL_ERROR)
  {
    EspUpdater::getInstance().reset();
  }
  if(EspUpdater::getInstance().getStatus() == IUpdater::UpdateStatus::COMPLETED) {
    STM_BootManager::toResetState();
    delay(400);
    ESP.restart();
  }

  IUpdater::HandleResult stm_handle_res = StmUpdater::getInstance().handle();
  if(stm_handle_res == IUpdater::FATAL_ERROR)
  {
    StmUpdater::getInstance().reset();
    Serial.end();
    Serial.begin(UART_BAUDRATE);
  }
  IUpdater::UpdateStatus stm_update_status = StmUpdater::getInstance().getStatus();
  switch (stm_update_status)
  {
  case IUpdater::UpdateStatus::FAILED:
    Serial.end();
    Serial.begin(UART_BAUDRATE);
    StmUpdater::getInstance().reset();
    STM32_OTA::getInstance().reset();
    break;
  case IUpdater::UpdateStatus::COMPLETED:
    Serial.end();
    Serial.begin(UART_BAUDRATE);
    StmUpdater::getInstance().reset();
    STM32_OTA::getInstance().finishUpdate();
    break;
  default:
    break;
  }
/*
  _resp_msg = NULL;
  if(_resp_msg != NULL)
  {
    int16_t txSize = mavlink_msg_to_send_buffer(mavSendBuffer, _resp_msg);
    if(_esp_now_mode)
    {
      EspNowSlave::getInstance()->addDataToSend(mavSendBuffer, txSize);
    }
    else
    {
      udpSender.sendToConnected(mavSendBuffer, txSize);
    }
  }
*/
  if(millis() >= _next_report_millis)
  {
    DEBUG_PRINTF("From boot %d\n", millis() / 1000);
    _next_report_millis += 1000 / REPORT_FREQ_HZ;
  }

  yield();
}

void MainController::mainRoutine()
{
  do 
  {
    _mainServiceRoutine();
  } while (_is_control_station_connected == false);
}

int MainController::getRC_Roll()
{
  return rcRoll;
}
int MainController::getRC_Pitch()
{
  return rcPitch;
}
int MainController::getRC_Throttle()
{
  return rcThrottle;
}
int MainController::getRC_Yaw()
{
  return rcYaw;
}

float MainController::getRoll()
{
  _enableAndWaitTelemetry(AttitudePacket, DEFAULT_ATTITUDE_TELEM_RATE);
  return currentRoll * RAD_TO_DEG;
}

float MainController::getPitch()
{
  _enableAndWaitTelemetry(AttitudePacket, DEFAULT_ATTITUDE_TELEM_RATE);
  return currentPitch * RAD_TO_DEG;
}

float MainController::getYaw()
{
  _enableAndWaitTelemetry(AttitudePacket, DEFAULT_ATTITUDE_TELEM_RATE);
  return currentYaw * RAD_TO_DEG;
}

float MainController::getRollRate()
{
  _enableAndWaitTelemetry(AttitudePacket, DEFAULT_ATTITUDE_TELEM_RATE);
  return _current_roll_rate * RAD_TO_DEG;
}

float MainController::getPitchRate()
{
  _enableAndWaitTelemetry(AttitudePacket, DEFAULT_ATTITUDE_TELEM_RATE);
  return _current_pitch_rate * RAD_TO_DEG;
}

float MainController::getYawRate()
{
  _enableAndWaitTelemetry(AttitudePacket, DEFAULT_ATTITUDE_TELEM_RATE);
  return _current_yaw_rate * RAD_TO_DEG;
}

float MainController::getPosX()
{
  _enableAndWaitTelemetry(LocalPosition, DEFAULT_LOCAL_POS_TELEM_RATE);
  return _current_pos_x;
}

float MainController::getPosY()
{
  _enableAndWaitTelemetry(LocalPosition, DEFAULT_LOCAL_POS_TELEM_RATE);
  return _current_pos_y;
}

float MainController::getPosZ()
{
  _enableAndWaitTelemetry(LocalPosition, DEFAULT_LOCAL_POS_TELEM_RATE);
  return _current_pos_z;
}

float MainController::getVelX()
{
  _enableAndWaitTelemetry(LocalPosition, DEFAULT_LOCAL_POS_TELEM_RATE);
  return _current_vel_x;
}

float MainController::getVelY()
{
  _enableAndWaitTelemetry(LocalPosition, DEFAULT_LOCAL_POS_TELEM_RATE);
  return _current_vel_y;
}

float MainController::getVelZ()
{
  _enableAndWaitTelemetry(LocalPosition, DEFAULT_LOCAL_POS_TELEM_RATE);
  return _current_vel_z;
}

float MainController::getAccelX()
{
  _enableAndWaitTelemetry(Standart9AxisIMU, DEFAULT_IMU_TELEM_RATE);
  return _current_accel_x;
}

float MainController::getAccelY()
{
  _enableAndWaitTelemetry(Standart9AxisIMU, DEFAULT_IMU_TELEM_RATE);
  return _current_accel_y;
}

float MainController::getAccelZ()
{
  _enableAndWaitTelemetry(Standart9AxisIMU, DEFAULT_IMU_TELEM_RATE);
  return _current_accel_z;
}

float MainController::getGyroX()
{
  _enableAndWaitTelemetry(Standart9AxisIMU, DEFAULT_IMU_TELEM_RATE);
  return _current_gyro_x * RAD_TO_DEG;
}

float MainController::getGyroY()
{
  _enableAndWaitTelemetry(Standart9AxisIMU, DEFAULT_IMU_TELEM_RATE);
  return _current_gyro_y * RAD_TO_DEG;
}

float MainController::getGyroZ()
{
  _enableAndWaitTelemetry(Standart9AxisIMU, DEFAULT_IMU_TELEM_RATE);
  return _current_gyro_z * RAD_TO_DEG;
}

int MainController::getAltitudeCentimeters()
{
  _enableAndWaitTelemetry(LocalPosition, DEFAULT_LOCAL_POS_TELEM_RATE);
  return _current_pos_z * 100;
}

float MainController::getDistanceSensorVal(uint16_t sensor_id)
{
  _enableAndWaitTelemetry(DistanceSensor, DEFAULT_DISTANCE_TELEM_RATE);
  if(sensor_id < MAX_SUPPORTED_DISTANCE_SENSORS_COUNT)
  {
    return _current_dist_sensor_val[sensor_id];
  }
  else
  {
    return -1;
  }
}

float MainController::getBatteryVoltage()
{
  return _battery_voltage;
}

void WiFiEvent(WiFiEvent_t event) {
#if DEBUG
  Serial.printf("[WiFi-event] event: %d\n", event);
#endif
  switch (event)
  {
    case WIFI_EVENT_STAMODE_CONNECTED:
#if DEBUG
      Serial.println("[ST]WiFi just connected");
#endif
      break;
    case WIFI_EVENT_STAMODE_DISCONNECTED:
#if DEBUG
      Serial.println("[ST]WiFi lost connection");
#endif
      break;
    case WIFI_EVENT_STAMODE_AUTHMODE_CHANGE:
#if DEBUG
      Serial.println("[ST]WiFi mode changed");
#endif
      break;
    case WIFI_EVENT_STAMODE_GOT_IP:
#if DEBUG
      Serial.print("[ST]WiFi connected, IP=");
      Serial.println(WiFi.localIP());
#endif
      break;
    case WIFI_EVENT_STAMODE_DHCP_TIMEOUT:
#if DEBUG
      Serial.println("[ST]WiFi DHCP TIMEOUT");
#endif
      break;
    case WIFI_EVENT_SOFTAPMODE_STACONNECTED:
#if DEBUG
      Serial.println("[AP] new client");
      printClientsStatus();
#endif
      break;
    case WIFI_EVENT_SOFTAPMODE_STADISCONNECTED:
#if DEBUG
      Serial.println("[AP]client disconnected");
#endif
      break;
    case WIFI_EVENT_SOFTAPMODE_PROBEREQRECVED:
#if DEBUG
      Serial.println("[AP]err on recieved request");
#endif
      break;
  }

}

}

