/*
    Copyright (C) 2019 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#ifndef MAIN_CONTROLLER_H
#define MAIN_CONTROLLER_H

#include <ESP8266WiFi.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include "ESPAsyncUDP.h"
#include "stm_boot_manager.h"
#include "stm_ota_udp_receiver.h"
#include "settings_manager.h"
#include "Color.h"
#include "UdpPacketSender.h"
#include "nanopix_config.h"
#include "RingBuffer.h"
#include "EspOTA.h"
#include <VL53L0X.h>
#include <VL53L1X.h>

extern "C" {
#include<user_interface.h>
}

namespace NanopixCore {

#define STM32_OTA_UDP_PORT 5700
#define STM32_OTA_SERVER_UDP_PORT 5705

#define CMD_LOOP_FREQ 19
#define CMD_UDP_RX_FREQ 20

#define MAX_SUPPORTED_SERVO_ID 20
#define MAX_SUPPORTED_MOTORS 20

void internal_init();

class MainController {
  /* Constrain values: */
  const float MIN_THROTTLE_VALUE = -1.0f;
  const float MAX_THROTTLE_VALUE = 1.0f;
  static const int MAX_YAW_RATE = 55; // deg/sec
  static const int MAX_PITCH_ROLL_SETPOINT_DEG = 35;
  static const int MAX_YAW_SETPOINT_DELTA_DEG = 360; 
  static const int MAX_ALTITUDE_SETPOINT_MM = 3000; 
  static const int MIN_ALTITUDE_SETPOINT_MM = 150; 
  static const int MIN_SERVO_ANGLE = 0;
  static const int MAX_SERVO_ANGLE = 180;
  static const int MAX_TELEMETRY_RATE_HZ = 300;
  static const int MIN_SERVO_NUMBER = 1;
  static const int MIN_SERVO_PWM_US = 0;
  static const int MAX_SERVO_PWM_US = 20000;
  const float MIN_YAW_SPEED_RAD = 0.1f;
  /* XYZ position limits: */
  const float MIN_ALTITUDE_SPEED = 0.1f; // meters/sec
  const float MAX_POS_Z_SPEED = 0.5f; // meters/sec 
  #define DEFAULT_POS_Z_MOVE_SPEED (0.2f) // meters/sec 
  const float MAX_ALTITUDE_OFFSET = 3.0f; // meters
  const float MAX_POS_XY_SPEED = 1.0f; // meters/sec
  const float MAX_POS_XY_OFFSET = 5.0f; // meters
  const float POS_CMD_TOLERANCE = 0.06f; // meters
  #define DEFAULT_POS_SPEED (0.4f) // meters/sec
  
  /* Timings and freq limits: */
  static const int OFFSET_ACCUMULATION_TIMEOUT_MS = 2000;
  static const int STATUS_REPORT_INTERVAL_MS = 1000;
  static const int ESP_OTA_PACKETS_MAX_TIMEOUT_MS = 3000;
  static const int RESET_SSID_PASS_SECONDS = 60 * 5;
  static const int MAX_SET_COLORS_FREQ_HZ = 10;
  static const int MAX_SET_MOTORS_FREQ_HZ = 500;
  static const int MAX_SET_YAW_FREQ_HZ = 50;
  static const int MAX_SEND_LOG_FREQ_HZ = 35;
  static const int MAX_SET_ALTITUDE_FREQ_HZ = 50;
  static const int MAX_SET_SERVO_FREQ_HZ = 300;
  static const int MAX_SET_POS_FREQ_HZ = 10;
  static const int MAX_COMMANDS_FREQ_HZ = 30;
  static const int ALTITUDE_CMD_TIMEOUT_SEC = 10; // TODO: needed more precise condition
  static const int POSITION_CMD_TIMEOUT_SEC = 30; // TODO: needed more precise condition
  static const int TAKEOFF_CMD_TIMEOUT_SEC = 10; // TODO: needed more precise condition, based on target altitude and takeoff speed
  static const int ARM_CMD_TIMEOUT_SEC = 3;
  static const int DEFAULT_TELEMETRY_WAITING_TIMEOUT_MS = 200;
  static const int TELEMETRY_ENABLE_RETRY_COUNT = 4;
  static const int DEFAULT_IMU_TELEM_RATE = 50;
  static const int DEFAULT_LOCAL_POS_TELEM_RATE = 35;
  static const int DEFAULT_DISTANCE_TELEM_RATE = 35;
  static const int DEFAULT_ATTITUDE_TELEM_RATE = 50;

  /* Buffers: */
  static const int PIX_RECEIVER_BUFFER_SIZE = 2048;
  static const int MAX_UART_READ_SIZE = 128;
  static const int MAVLINK_TX_BUFFER_SIZE = 256;
  static const int UART_BRIDGE_BUFFER_SIZE = 1024;
  static const int UDP_RX_BUFFER = 256;
  static const int MAX_TRACKED_TELEMETRY_COUNT = 5;
  static const int MAX_SUPPORTED_DISTANCE_SENSORS_COUNT = 11; // main sensor + 10 slaves

  /* Values: */
  static const int DEFAULT_YAW_SPEED = 35;
  static const int MIN_CLAW_SERVO_ANGLE = 60;
  static const int MAX_CLAW_SERVO_ANGLE = 180;
  static const int MAX_CLAW_COUNT = 2;

public:
  enum TelemetryPacketType {Standart9AxisIMU = 2, AttitudePacket, DistanceSensor, LocalPosition};
  enum AttitudeControlMode {OverrideMode = 1, FusionMode};

  
private:
  
  volatile bool _is_telemetry_enabled[MAX_TRACKED_TELEMETRY_COUNT];
  volatile bool _is_telemetry_packet_received[MAX_TRACKED_TELEMETRY_COUNT];

  RingBuffer<uint8_t, UART_BRIDGE_BUFFER_SIZE> _uart_bridge_buffer;
  mavlink_message_t _rx_mav_msg, _uart_rx_mav_msg;
  mavlink_status_t _mav_status, _uart_mav_status;
  uint8_t mavSendBuffer[MAVLINK_TX_BUFFER_SIZE];
  AttitudeControlMode attitudeControlMode;
  UdpPacketSender udpSender;
  AsyncUDP Udp;
  int localUdpPort = 14555;

  mavlink_message_t _tx_mav_msg;
  EspOTA _esp_ota_manager;
  mavlink_message_t * _resp_msg;

  /* For user/pass reset: */
  bool isSomeConnectedBefore = false;
  int noConnectSeconds = 0;
  /* -------------------- */

  volatile bool _is_mission_running;
  volatile bool _async_udp_blocked = false;
  volatile bool _is_control_station_connected = false;
  uint32_t _next_flush_stuck_commands_ms;
  unsigned long currentMillis;
  unsigned long prevMillis = 0;
  uint8_t pixReceiverBuffer[PIX_RECEIVER_BUFFER_SIZE];
  uint8_t udpRxBuffer[UDP_RX_BUFFER];
  SettingsManager & settingsManager;
  bool isManualControlOverrided = false;
  float pitchControl, rollControl, throttleControl, yawControl;
  bool motorsArmCommandReceived = false;
  float _current_pos_x;
  float _current_pos_y; 
  volatile float _current_pos_z; 
  volatile float _current_vel_x;
  volatile float _current_vel_y; 
  volatile float _current_vel_z; 
  volatile float _battery_voltage;
  volatile float _current_dist_sensor_val[MAX_SUPPORTED_DISTANCE_SENSORS_COUNT];
  volatile float currentRoll, currentPitch, currentYaw;
  volatile float _current_roll_rate, _current_pitch_rate, _current_yaw_rate;
  volatile int rcRoll, rcPitch, rcThrottle, rcYaw;
  volatile bool missionStartCmdReceived;
  int wifiChannel;
  static MainController * _instance;
  volatile bool _esp_now_mode = false;
  uint32_t _next_report_millis;
  volatile float _current_accel_x, _current_accel_y, _current_accel_z;
  volatile float _current_gyro_x, _current_gyro_y, _current_gyro_z;
  uint32_t _last_set_front_color_ms, _last_set_back_color_ms;
  uint32_t _last_set_motor_ms[MAX_SUPPORTED_MOTORS];
  uint32_t _last_set_yaw_ms;
  float _yaw_relative_accumulator;
  uint32_t _last_yaw_accumulation_ms;
  uint32_t _last_send_log_ms;
  uint32_t _last_set_altitude_ms;
  float _altitude_relative_accumulator;
  uint32_t _last_altitude_accumulation_ms;
  uint32_t _last_set_servo_ms[MAX_SUPPORTED_SERVO_ID];
  uint32_t _last_set_target_pos_ms;
  float _pos_offset_accumulator[3];
  uint32_t _last_pos_accumulation_ms;
  uint32_t _last_arming_ms, _last_disarming_ms;
  uint32_t _last_takeoff_cmd_ms, _last_land_cmd_ms;
  uint32_t _last_reboot_cmd_ms;
  uint32_t _last_headfree_cmd_ms;
  volatile bool _yaw_cmd_ack_received;
  volatile bool _altitude_cmd_ack_received;
  volatile bool _position_cmd_ack_received;
  volatile bool _takeoff_land_ack_received;
  volatile bool _arm_cmd_ack_received;
  volatile bool _is_headfree_enabled;
  volatile bool _stored_headfree_state;
  volatile bool _is_motionless;
  volatile bool _is_cmd_from_interrupt;

  volatile bool _is_front_led_cmd_stuck;
  volatile bool _is_back_led_cmd_stuck;
  volatile bool _is_set_motor_stuck[MAX_SUPPORTED_MOTORS];
  volatile bool _is_set_absolute_yaw_stuck, _is_set_relative_yaw_stuck;
  ///volatile bool _last_send_log_ms;  // TODO add queue for logs
  volatile bool _is_absolute_altitude_stuck, _is_relative_altitude_stuck;
  volatile bool _is_set_servo_angle_stuck[MAX_SUPPORTED_SERVO_ID];
  volatile bool _is_set_servo_pwm_stuck[MAX_SUPPORTED_SERVO_ID];
  volatile bool _is_set_target_pos_stuck, _is_offset_pos_stuck;
  volatile bool _is_arming_stuck, _is_disarming_stuck;
  volatile bool _is_takeoff_cmd_stuck, _is_land_cmd_stuck;
  volatile bool _is_reboot_cmd_stuck;
  volatile bool _is_headfree_cmd_stuck;

  Color _stuck_front_led_color, _stuck_back_led_color;
  volatile float _stuck_set_motor_value[MAX_SUPPORTED_MOTORS];
  volatile float _stuck_yaw_value[2];
  volatile float _stuck_altitude_value[2];
  volatile int _stuck_servo_value[MAX_SUPPORTED_SERVO_ID];
  volatile float _stuck_target_pos[6];
  volatile bool _stuck_arm_disarm_value;
  volatile bool _stuck_takeoff_land_value;
  volatile bool _stuck_headfree_value;

  enum RangeSensorType { UNDEF_TYPE = 0, VL53L0X_TYPE, VL53L1X_TYPE };
  VL53L0X _range_sensor;
  VL53L1X _long_range_sensor;
  uint32_t _last_long_range_read_ts;
  uint32_t _last_short_range_read_ts;
  uint32_t _is_range_sensor_initialized;
  RangeSensorType _range_sensor_type;

  int _claw_angle[MAX_CLAW_COUNT];
  static int _claw_servo_ids[MAX_CLAW_COUNT];


  /* -------------------- Friend functions: -------------------------------- */
  friend void espNowOnReceivedHandler(uint8_t * data, uint8_t len);
  friend void retransmitTimerCallback(void *pArg);
  friend void udpOnPacketHandler(AsyncUDPPacket packet);
  friend void WiFiEvent(WiFiEvent_t event);
  friend void internal_init();

  /* -------------------- Private methods: -------------------------------- */
  void _setParam(const char* name, float value);
  void _sendMavCommandInt(uint8_t msgType, 
                            float param1 = 0, float param2 = 0, 
                            float param3 = 0, float param4 = 0, 
                            float param5 = 0, float param6 = 0, 
                            float param7 = 0);

  void _switchTakeoffLand(bool takeoff, bool is_blocking);
  void _switchArmDisarm(bool armState, bool is_blocking);
  int _sendMavlinkPacket(mavlink_message_t* msg, bool multicast = false);
  void _openAsyncUDP();
  void _closeAsyncUDP();
  void _mainServiceRoutine();

  void _setPosition(float x, float y, float z, 
                    float vel_x, float vel_y, float vel_z,
                    bool is_relative, bool is_blocking);

  void _setHeadfree(bool is_enabled);
  void _setAltitude(float meters, float max_speed, bool is_relative, bool is_blocking);
  void _setYaw(float yaw_deg, float speed_deg_sec, bool is_relative, bool is_blocking);
  void _retransmitFromWireless(uint8_t * data, uint8_t length);
  void _grabSerialAndRetransmit(); // extract data from packets and retransmit it to GS

  void _tryFlushStuckCommands(); // flush rejected by frequency messages from ESP to FC
  
  bool _enableAndWaitTelemetry(TelemetryPacketType stream_id, uint16_t stream_rate,
                               uint32_t timeout_ms = DEFAULT_TELEMETRY_WAITING_TIMEOUT_MS);

  void _handleEspOta(mavlink_message_t * mavmsg);
  void _missionPreparingReset();

  /* VL53L0X laser range sensor: */
  bool _initializeShortRangeSensor();
  float _readShortRangeSensorValue(); // returns value in centimeters

  /* VL53L1X laser range sensor: */
  bool _initializeLongRangeSensor();
  float _readLongRangeSensorValue(); // returns value in centimeters

  MainController();
  MainController(MainController const&) = delete;
  MainController& operator=(MainController const&) = delete;
public:
  /* Service: */
  ~MainController();
  static MainController & getInstance();
  void initializeSystem();
  void mainRoutine();  
  /* ------------------------------------------------------------------------------- */

  /* Sensors API: */

  /* IR line sensor: */
  void initializeLineSensor(int digital_pin = 2, bool with_power = true);
  bool readLineSensorValue(int digital_pin = 2); // true - surface detected, false - no surface

  /* Laser range sensor: */
  float readExternalRangeSensorValue(); // returns value in centimeters

  /* Manual control: */
  void setThrottle(float throttle); // [-1...1] | 0 <-> hover throttle
  void setPitch(float deg_angle, bool pitch_only = true); // degrees angle
  void setRoll(float deg_angle, bool roll_only = true);  // degrees angle
  void setYawSpeed(float speed_deg_sec); // degrees per second speed, see MAX_YAW_RATE as limit

  /* Setters: */
  void setYawDegrees(float target_yaw_deg, 
                     float speed_deg_sec = DEFAULT_YAW_SPEED, 
                     bool is_blocking = true); // set absolute yaw angle, target_yaw_deg in range [-PI, PI]

  void rotateYawDegrees(float delta_yaw_deg, // delta_yaw_deg in range [-PI, PI]
                        float speed_deg_sec = DEFAULT_YAW_SPEED, // delta_yaw_deg in range [6, 55] deg/sec 
                        bool is_blocking = true); // relative yaw moving with speed='speed_deg_sec'

  void setTargetAltitude(float meters, float max_speed = DEFAULT_POS_Z_MOVE_SPEED, 
                         bool is_blocking = true); // meters in range [0.15 ... 3.0]

  void offsetTargetAltitude(float delta_meters, float max_speed = DEFAULT_POS_Z_MOVE_SPEED, 
                              bool is_blocking = true); // delta_meters in range [-3.0 ... 3.0]

  void setTargetPos(float pos_x, float pos_y, float pos_z, 
                    float vel_x = DEFAULT_POS_SPEED, float vel_y = DEFAULT_POS_SPEED,
                    float vel_z = DEFAULT_POS_Z_MOVE_SPEED, bool is_blocking = true);

  void setTargetPos(float pos_x, float pos_y, bool is_blocking);

  void offsetTargetPos(float pos_dx, float pos_dy, float pos_dz, 
                       float vel_x = DEFAULT_POS_SPEED, float vel_y = DEFAULT_POS_SPEED, 
                       float vel_z = DEFAULT_POS_Z_MOVE_SPEED, bool is_blocking = true);

  void offsetTargetPos(float pos_dx, float pos_dy, bool is_blocking = true);

  void moveForward(float dist, float speed = DEFAULT_POS_SPEED, bool is_blocking = true);
  void moveBackward(float dist, float speed = DEFAULT_POS_SPEED, bool is_blocking = true);
  void moveRight(float dist, float speed = DEFAULT_POS_SPEED, bool is_blocking = true);
  void moveLeft(float dist, float speed = DEFAULT_POS_SPEED, bool is_blocking = true);
  
  void setAttitudeControlMode(AttitudeControlMode attitudeCtrlMode);
  // setServoAngle: servo_number in range [1...lim], lim depends on board (see board specs), limited by MAX_SUPPORTED_SERVO_ID, angle_deg [0...180]
  void setServoAngle(int servo_number, int angle_deg); 
  void setServoPWM(int servo_number, int pwm_value_us); // pwm_value_us in range [0..20000] microseconds
  void disableServo(int servo_number); // disable servo control (disable PWM control)
  void disableAllServo(); // disable servo control for all servo motors

  // setMotorSpeed: 'motorNumber' from 1 to 4 (as on-board numbers), 'speedPercent' from 0.0 to 100.0
  void setMotorSpeed(uint8_t motor_number, float speed_percent); 

  // LEDs setters:
  void setLEDsColors(uint8_t forwardR, uint8_t forwardG, uint8_t forwardB, 
                     uint8_t backwardR, uint8_t backwardG, uint8_t backwardB);
  void setLEDsColors(Color & forwardColor, Color & backwardColor);
  void setFrontLEDColor(uint8_t forwardR, uint8_t forwardG, uint8_t forwardB);
  void setBackLEDColor(uint8_t backwardR, uint8_t backwardG, uint8_t backwardB);
  void setFrontLEDColor(Color & forwardColor);
  void setBackLEDColor(Color & backwardColor);
  /* ------------------------------------------------------------------------------- */

  /* Getters: */
  // Position:
  float getPosX();
  float getPosY();
  float getPosZ();
  int getAltitudeCentimeters();
  float getVelX();
  float getVelY();
  float getVelZ();
  float getDistanceSensorVal(uint16_t sensor_id = 0);
  // IMU:
  float getAccelX();
  float getAccelY();
  float getAccelZ();
  float getGyroX(); // deg/sec
  float getGyroY(); // deg/sec
  float getGyroZ(); // deg/sec
  // Attitude:
  float getRoll(); // degrees
  float getPitch(); // degrees
  float getYaw(); // degrees
  float getRollRate(); // deg/sec
  float getPitchRate(); // deg/sec
  float getYawRate(); // deg/sec
  // Power:
  float getBatteryVoltage();
  // RC control:
  int getRC_Roll();
  int getRC_Pitch();
  int getRC_Throttle();
  int getRC_Yaw();
  /* ------------------------------------------------------------------------------- */

  /* Commands and requests: */
  void armMotors(bool is_blocking = true);
  void disarmMotors();
  void takeoff(bool is_blocking = true);
  void land(bool is_blocking = true);
  void enableHeadfree();
  void disableHeadfree();
  void rebootFlightController();
  void requestTelemetryDataStream(TelemetryPacketType streamType, int streamRateHz, bool startStopFlag);
  /* ------------------------------------------------------------------------------- */

  /* Logging */
  int sendLogMsg(const char* str, bool multicast = false);
  int sendLogMsg(int value, bool multicast = false);
  int sendLogMsg(float value, bool multicast = false);
  /* ------------------------------------------------------------------------------- */

  /* Checks: */
  bool resetMissionTrigger();
  bool isMotionless();
  bool isMissionTriggered();
  bool isArmCommandReceived();
  bool isHeadfreeEnabled();
  /* ------------------------------------------------------------------------------- */

  void openClaw(int index = 0);
  void closeClaw(int index = 0);
  void setClaw(int percent, int index = 0); // percent = 0 -> fully closed, percent = 100 -> fully opened

  /* UART bridge: ESP UART (you here) <-> STM32 UART (transparent bridge) <-> EXTERNAL DEVICE */
  /* Default baudrate: 115200 (can't be changed now) */
  void sendToUART(const uint8_t * data, uint16_t count);
  size_t availableBytesOnUART();
  uint16_t readFromUART(uint8_t * readBuffer, uint16_t readCount);
  /* ------------------------------------------------------------------------------- */
};

}

#endif
