/*
    Copyright (C) 2019 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#include "PacketsFifo.h"
#include "string.h"
#include <ESP8266WiFi.h>

namespace NanopixCore {

template <size_t MaxSize>
PacketsFifo<MaxSize>::PacketsFifo()
{
    _first_packet_id = 0;
    _packets_count = 0;
}

template <size_t MaxSize>
bool PacketsFifo<MaxSize>::pushPacket(uint8_t * data, size_t length)
{
    if(data == NULL || length <= 0)
    {
      return false;
    }
    noInterrupts();
    if((_packets_count == MaxSize) || (length > PacketInfo::MAX_PACKET_SIZE))
    {
      return false;
    }
    size_t free_packet_id = (_first_packet_id + _packets_count) % MaxSize;
    _packets[free_packet_id].length = length;
    ++_packets_count;
    interrupts();
    memcpy(_packets[free_packet_id].data, data, length);
    return true;
}

template <size_t MaxSize>
bool PacketsFifo<MaxSize>::peekPacket(uint8_t ** data, size_t & length)
{
    if(_packets_count == 0)
    {
      return false;
    }
    *data = _packets[_first_packet_id].data;
    length = _packets[_first_packet_id].length;
    return true;
}

template <size_t MaxSize>
bool PacketsFifo<MaxSize>::pullPacket(uint8_t ** data, size_t & length)
{
    if(_packets_count == 0)
    {
      return false;
    }
    *data = _packets[_first_packet_id].data;
    length = _packets[_first_packet_id].length;
    noInterrupts();
    _first_packet_id = (_first_packet_id + 1) % MaxSize;
    --_packets_count;
    interrupts();
    return true;
}

template <size_t MaxSize>
size_t PacketsFifo<MaxSize>::count() const
{
    return _packets_count;
}

}

