/*
    Copyright (C) 2017 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#include "STM32_OTA.h"
#include "stm_boot_manager.h"
#include "platform_hal.h"
#include "stm_ota_udp_receiver.h"
#include "nanopix_config.h"
//#include "utils.h"

namespace NanopixCore {

#define BOOTLOADER_GET_VERS_AND_RP 0x02 // crc: 0xFD
#define BOOTLOADER_WRITE_COMMAND 0x31 
#define BOOTLOADER_READ_COMMAND 0x11
#define BOOTLOADER_ERASE_COMMAND 0x44 //bc
#define BOOTLOADER_WRITE_UNPROTECT_COMMAND 0x73 //8c
#define BOOTLOADER_READOUT_UNPROTECT_COMMAND 0x92 // 6d

#define WAIT_ACK()\
        {\
            ack_status = waitBootloaderAck(); \
            if (ack_status != BOOTLOADER_ACK_OK) \
                goto out; \
        }

#define WAIT_ACK_TIMEOUT(timeout)\
        {\
            ack_status = waitBootloaderAck(timeout); \
            if (ack_status != BOOTLOADER_ACK_OK) \
                goto out; \
        }

STM32_OTA * STM32_OTA::_instance = NULL;

STM32_OTA & STM32_OTA::getInstance()
{
  if(_instance == NULL)
  {
    _instance = new STM32_OTA();
  }
  return *_instance;
}

 bool STM32_OTA::isOTA_Updating() {
    return isOTA_Update;
  }

  STM32_OTA::STM32_OTA()
  {
    reset();
  }

  bool STM32_OTA::begin(uint32_t ota_fw_size, bool mavlink_mode) {
    isOTA_Update = false;
    _ota_fw_size = ota_fw_size;
    bool is_size_accepted = (ota_fw_size < MAX_STM_FIRMWARE_SIZE);
    memset(_firmware_md5, 0, FIRMWARE_MD5_LEN);
    _is_md5_set = false;
    _md5.begin();
    _flashed_bytes_count = 0;
    _mavlink_mode = mavlink_mode;
    return is_size_accepted;
  }

  void STM32_OTA::setMD5Bytes(uint8_t * md5_bytes) // 16 bytes len
  {
    if(md5_bytes != NULL)
    {
      memcpy(_firmware_md5, md5_bytes, FIRMWARE_MD5_LEN);
      _is_md5_set = true;
    }
  }

  void STM32_OTA::startUpdate()
  {
    startAddrIsInit = false;
    startAddrHighMSB = 0x08;
    startAddrHighLSB = 0x00;
  }

  STM32_OTA::BootloaderAckStatus STM32_OTA::enterToBOOT()
  {
    //if(!isOTA_Update)
    {
        serial_link.end();
        serial_link.begin(256000);
        while(serial_link.available())
          serial_link.read();
    }
    isOTA_Update = true;
    uint8_t newByte;
    uint8_t tryCount = 5;
    for(int i = 0; i < tryCount; i++)
    {
      STM_BootManager::toBootMode();
      delayMs(70);     
      while(serial_link.available())
          serial_link.read();   
      serial_link.write(0x7F);
      int timeout = 100000;
      while(!serial_link.available() && timeout--);
      if(timeout == 0)
      {
        if(!_mavlink_mode) {
          STM_OTA_UDP_Receiver::sendLog("Enter to BOOT mode ACK timeout" );
          STM_OTA_UDP_Receiver::sendResponce(0x00);
        }
        return BOOTLOADER_ACK_TIMEOUT;
      }
      newByte = serial_link.read();
      if(newByte == ACK_TOKEN_BYTE)
        break;
      if(i < tryCount - 1) {
        //STM_BootManager::reset();
        delayMs(100);
        if(!_mavlink_mode) {
          STM_OTA_UDP_Receiver::sendLog("Enter to BOOT failed, try again..." );
        }
      } 
    }

    if(!_mavlink_mode) {
      STM_OTA_UDP_Receiver::sendResponce(newByte);
    }
    if(newByte == ACK_TOKEN_BYTE)
        return BOOTLOADER_ACK_OK;
    else
        return BOOTLOADER_ACK_FAIL;
  }
  
  void STM32_OTA::finishUpdate()
  {
    isOTA_Update = false;
    STM_BootManager::reset();
    reset();
    if(!_mavlink_mode) {
      STM_OTA_UDP_Receiver::sendLog("Finish update!");
    }
  }

  void STM32_OTA::reset()
  {
    isOTA_Update = false;
    startAddrIsInit = false;
    startAddrHighMSB = 0x08;
    startAddrHighLSB = 0x00;
    memset(_firmware_md5, 0 , FIRMWARE_MD5_LEN);
    _is_md5_set = false;
    _current_addr = 0;
    _ota_fw_size = 0;
    _flashed_bytes_count = 0;
    _mavlink_mode = false;
  }

  void STM32_OTA::_parseSettingsRecord(uint8_t* packet_data, size_t & out_flashed_bytes)
  {
    int payload_len = packet_data[0] + 1;
    
    _md5.add(&(packet_data[0]), payload_len + 5);
    _flashed_bytes_count += payload_len + 5;

    if(packet_data[3] == 4)
    {
      startAddrHighMSB = packet_data[4];
      startAddrHighLSB = packet_data[5];
      startAddrIsInit = true;
    }
    out_flashed_bytes = payload_len + 5;
  }

  void STM32_OTA::_sendAddress(uint8_t addrMSB, uint8_t addrLSB)
  {
    uint8_t addrXOR = startAddrHighMSB ^ startAddrHighLSB ^ addrMSB ^ addrLSB;
    serial_link.write(startAddrHighMSB);
    serial_link.write(startAddrHighLSB);
    serial_link.write(addrMSB);
    serial_link.write(addrLSB);
    serial_link.write(addrXOR);
  }

  uint32_t STM32_OTA::progress()
  {
    return _flashed_bytes_count;
  }

  bool STM32_OTA::end()
  {
    _md5.calculate();
    uint8_t computed_md5_bytes[FIRMWARE_MD5_LEN];
    _md5.getBytes(computed_md5_bytes);
    //DEBUG_PRINTF("STM mav OTA MD5 bytes: %d\n", _md5_bytes_counter);
    return (memcmp(_firmware_md5, computed_md5_bytes, FIRMWARE_MD5_LEN) == 0);
  }

  bool STM32_OTA::isFinished()
  {
    return (progress() == _ota_fw_size);
  }

  STM32_OTA::BootloaderAckStatus STM32_OTA::write(uint8_t* packet_data, size_t data_size, 
                                                  size_t & out_flashed_bytes)
  {
    STM32_OTA::BootloaderAckStatus ack_status = BOOTLOADER_ACK_OK;
    
    out_flashed_bytes = 0;
    if(data_size < 4) 
    {
      return WRONG_PACKET;
    }
    if(packet_data[3] != 0)
    {
      _parseSettingsRecord(packet_data, out_flashed_bytes);
      return BOOTLOADER_ACK_OK;
    }
    else
    {
      if(startAddrIsInit == false)
      {
        if(!_mavlink_mode) {
          STM_OTA_UDP_Receiver::sendLog("ERROR: start addr not init!");
        }
        return STATUS_UNINITIALIZED;
      }

      int payload_len = packet_data[0] + 1;
      uint8_t addrMSB = packet_data[1];
      uint8_t addrLSB = packet_data[2];
      uint8_t checksum = payload_len - 1;

      if(data_size < payload_len + 5) 
      {
        return WRONG_PACKET;
      }

      for(int i = 0; i < payload_len; i++)
      {
        checksum = checksum ^ packet_data[4 + i];
      }

      _md5.add(&(packet_data[0]), payload_len + 5);

      _current_addr = (((uint16_t)addrMSB) << 8) | addrLSB;

      serial_link.write(BOOTLOADER_WRITE_COMMAND);
      serial_link.write(BOOTLOADER_WRITE_COMMAND ^ 0xFF);

      WAIT_ACK();
        
      _sendAddress(addrMSB, addrLSB);

      WAIT_ACK();

      serial_link.write(payload_len-1);
      serial_link.write(&(packet_data[4]), payload_len);
      serial_link.write(checksum);

      WAIT_ACK();

      if(ack_status == BOOTLOADER_ACK_OK)
      {
        out_flashed_bytes = payload_len + 5;
        _flashed_bytes_count += payload_len + 5;
      }
      else
      {
        out_flashed_bytes = 0;
      }
    }
out:
    return ack_status;
  }

  STM32_OTA::BootloaderAckStatus STM32_OTA::verify(uint8_t* packet_data)
  {
    STM32_OTA::BootloaderAckStatus ack_status = BOOTLOADER_ACK_OK;
    bool dataIsCorrect = true;
    uint8_t recvData[270]; // TODO: check max block size
    int timeout;
    uint8_t dataSize = packet_data[0];
    uint8_t addrMSB = packet_data[1];
    uint8_t addrLSB = packet_data[2];

    serial_link.write(BOOTLOADER_READ_COMMAND);
    serial_link.write(BOOTLOADER_READ_COMMAND ^ 0xFF);

    WAIT_ACK();
      
    _sendAddress(addrMSB, addrLSB);

    WAIT_ACK();

    serial_link.write(dataSize-1);
    serial_link.write((dataSize-1) ^ 0xFF);

    WAIT_ACK();

    
    for(int i = 0; i < dataSize; i++)
    {
      timeout = 100000;
      while(serial_link.available() == 0 && timeout--);
      recvData[i] = serial_link.read();
      if(recvData[i] != packet_data[4+i] || timeout == 0)
      {
        dataIsCorrect = false;
        break;
      }
    }

    if(dataIsCorrect)
    {
      //sendResponce(ACK_TOKEN_BYTE);
      return BOOTLOADER_ACK_OK;
    } 
    else 
    {
      if(!_mavlink_mode) {
        STM_OTA_UDP_Receiver::sendResponce(0x00);
      }
      return BOOTLOADER_ACK_FAIL;
    }
out:
    return ack_status;
  }

  STM32_OTA::BootloaderAckStatus STM32_OTA::globalErase()
  {
    STM32_OTA::BootloaderAckStatus ack_status = BOOTLOADER_ACK_OK;

    serial_link.write(BOOTLOADER_ERASE_COMMAND);
    serial_link.write(BOOTLOADER_ERASE_COMMAND ^ 0xFF);

    WAIT_ACK();

    serial_link.write(0xFF);
    serial_link.write(0x00);

    WAIT_ACK_TIMEOUT(2000000);
out:
    return ack_status;
  }

  STM32_OTA::BootloaderAckStatus STM32_OTA::globalEraseEx()
  {
    STM32_OTA::BootloaderAckStatus ack_status = BOOTLOADER_ACK_OK;

    serial_link.write(BOOTLOADER_ERASE_COMMAND);
    serial_link.write(BOOTLOADER_ERASE_COMMAND ^ 0xFF);

    WAIT_ACK();

    serial_link.write(0xFF);
    serial_link.write(0xFF);
    serial_link.write(0x00);

    WAIT_ACK_TIMEOUT(2000000);
out:
    return ack_status;
  }

  STM32_OTA::BootloaderAckStatus STM32_OTA::setReadoutUnprotect()
  {
    STM32_OTA::BootloaderAckStatus ack_status = BOOTLOADER_ACK_OK;

    serial_link.write(BOOTLOADER_READOUT_UNPROTECT_COMMAND);
    serial_link.write(BOOTLOADER_READOUT_UNPROTECT_COMMAND ^ 0xFF);

    WAIT_ACK();
    WAIT_ACK_TIMEOUT(2000000);
out:
    return ack_status;
  }

  STM32_OTA::BootloaderAckStatus STM32_OTA::setWriteUnprotect()
  {
    STM32_OTA::BootloaderAckStatus ack_status = BOOTLOADER_ACK_OK;

    serial_link.write(BOOTLOADER_WRITE_UNPROTECT_COMMAND);
    serial_link.write(BOOTLOADER_WRITE_UNPROTECT_COMMAND ^ 0xFF);

    WAIT_ACK();

out:
    return ack_status;
  }
  
  STM32_OTA::BootloaderAckStatus STM32_OTA::waitBootloaderAck(int timeout)
  {
    while(!serial_link.available() && timeout--);
    uint8_t newByte = serial_link.read();
    if(timeout == 0) 
    {  
      if(!_mavlink_mode) {
        STM_OTA_UDP_Receiver::sendResponce(0);
      }
      return BOOTLOADER_ACK_TIMEOUT;
      //return false;
    }
    if(newByte != ACK_TOKEN_BYTE)
    {
      if(!_mavlink_mode) {
        STM_OTA_UDP_Receiver::sendResponce(newByte);
      }
      return BOOTLOADER_ACK_FAIL;
      //return false;
    }
    return BOOTLOADER_ACK_OK;
  }

}
