/*
    Copyright (C) 2020 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#include "StmUpdater.h"
#include "STM32_OTA.h"
#include "stdlib.h"
#include "nanopix_config.h"
#include "utils.h"
#include "stm_boot_manager.h"
#include "platform_hal.h"

namespace NanopixCore {

StmUpdater * StmUpdater::_instance = NULL;

StmUpdater::StmUpdater()
{
  _temp_buffer = NULL;
  _buffer_processed_len = 0;
  _boot_entered = false;
  _retry_write_counter = 0;
  _retry_boot_counter = 0;
  _retry_erase_counter = 0;
  _last_new_data_ts = 0;
  _start_waiting_ts = 0;
  _last_handle_result = HandleResult::HANDLE_COMPLETED;
  _custom_fatal_code = NO_ERROR;
  _status = IUpdater::NOT_STARTED;
}

StmUpdater & StmUpdater::getInstance()
{
    if(_instance == NULL)
    {
        _instance = new StmUpdater();
    }
    return *_instance;
}

bool StmUpdater::begin(uint32_t firmware_size, uint8_t * md5_bytes)
{
  DEBUG_PRINTF("STM OTA begin, fw size: %d\n", (int)firmware_size);
  reset();
  STM32_OTA::getInstance().begin(firmware_size, true);
  STM32_OTA::getInstance().setMD5Bytes(md5_bytes);
  _boot_entered = false;
  _retry_write_counter = 0;
  _retry_boot_counter = 0;
  _retry_erase_counter = 0;
  _last_new_data_ts = 0;
  _start_waiting_ts = 0;
  _last_handle_result = HandleResult::HANDLE_COMPLETED;
  _custom_fatal_code = NO_ERROR;
  _status = IUpdater::IN_PROCESS;
  return true;
}


IUpdater::UpdateStatus StmUpdater::getStatus() const
{
  return _status;
}

void StmUpdater::end()
{
    bool md5_ok = STM32_OTA::getInstance().end();
    if(md5_ok)
    {
      DEBUG_PRINTF("STM OTA end COMPLETED\n");
      _status = COMPLETED;
    }
    else
    {
      DEBUG_PRINTF("STM OTA end FAILED\n");
      _status = FAILED;
    }
}

bool StmUpdater::isFwSizeValid(int32_t firmware_size)
{
  if(firmware_size <= 0 || firmware_size > MAX_STM_FIRMWARE_SIZE)
    return false;
  //STM32_OTA::getInstance().finishUpdate();
  return STM32_OTA::getInstance().begin(firmware_size);
}

bool StmUpdater::isBufferCompleted()
{
  if(_temp_buffer == NULL)
    return true;
  return (_buffer_processed_len == _temp_buffer->length);
}

void StmUpdater::setBuffer(FwBuffer * temp_buffer)
{
  _temp_buffer = temp_buffer;
  _buffer_processed_len = 0;
}

void StmUpdater::reset()
{
  //STM32_OTA::getInstance().finishUpdate();
  _buffer_processed_len = 0;
  _temp_buffer = NULL;
  _status = IUpdater::NOT_STARTED;
  _boot_entered = false;
  _retry_write_counter = 0;
  _retry_boot_counter = 0;
  _retry_erase_counter = 0;
  _start_waiting_ts = 0;
  _last_handle_result = HandleResult::HANDLE_COMPLETED;
  _custom_fatal_code = NO_ERROR;
  _last_new_data_ts = 0;
}

int StmUpdater::getLastFatalCode()
{
  return (int)_custom_fatal_code;
}

IUpdater::HandleResult StmUpdater::getLastHandleResult()
{
  return _last_handle_result;
}

bool StmUpdater::_tryEnterToBoot()
{
  STM32_OTA::BootloaderAckStatus ack_status = STM32_OTA::getInstance().enterToBOOT();
  if(ack_status == STM32_OTA::BOOTLOADER_ACK_OK)
  {
    ack_status = STM32_OTA::getInstance().globalEraseEx();
    if(ack_status == STM32_OTA::BOOTLOADER_ACK_OK)
    {
      STM32_OTA::getInstance().startUpdate();
      DEBUG_PRINTF("STM OTA started\n");
      _last_handle_result = HandleResult::HANDLE_COMPLETED;
      _custom_fatal_code = NO_ERROR;
      return true;
    }
    else
    {
      // Global erase FAILED:
      if(_retry_erase_counter < MAX_RETRY_ERASE_COUNT) 
      {
        _retry_erase_counter++;
      }
      else
      {
        _setFatalResult(ERASE_FAILED);
        _retry_erase_counter = 0;
      }
    }
  }
  else
  {
    // STM enter to boot mode FAILED:
    if(_retry_boot_counter < MAX_RETRY_BOOT_COUNT) 
    {
      _retry_boot_counter++;
    }
    else
    {
        _setFatalResult(ENTER_TO_BOOT_FAILED);
        _retry_boot_counter = 0;
    }
  }
  return false;
}

void StmUpdater::_tryProcessBufferData()
{
  _custom_fatal_code = NO_ERROR;
  if(_buffer_processed_len < _temp_buffer->length)
  {
    _last_new_data_ts = systemMillis();
    size_t completed_bytes = 0;
    STM32_OTA::BootloaderAckStatus ack_status = STM32_OTA::getInstance().write(
                        &(_temp_buffer->data[_buffer_processed_len]),
                        _temp_buffer->length - _buffer_processed_len, 
                        completed_bytes);
    _buffer_processed_len += completed_bytes;
    if(ack_status == STM32_OTA::BOOTLOADER_ACK_OK)
    {
      _retry_write_counter = 0;
      _last_handle_result = HandleResult::HANDLE_COMPLETED;
    }
    else
    {
      if(_retry_write_counter < MAX_RETRY_WRITE_COUNT)
      {
        ++_retry_write_counter;
        _last_handle_result = HandleResult::ERROR_RETRY_BLOCK;
      }
      else
      {
        _setFatalResult(WRITE_FAILED);
      }
    }
  }
  else
  {
    if(_buffer_processed_len == _temp_buffer->length)
    {
      _last_handle_result = HandleResult::NO_NEW_DATA;
    }
    else
    {
      _setFatalResult(WRONG_PACKET);
    }
  }
}

void StmUpdater::_setFatalResult(int fatal_code)
{
  _custom_fatal_code = fatal_code;
  _last_handle_result = FATAL_ERROR;
}

IUpdater::HandleResult StmUpdater::handle()
{
  if(_status == IUpdater::IN_PROCESS)
  {
    if(STM32_OTA::getInstance().isFinished())
    {
      _status = IUpdater::FINISHED;
    }

    if(!_boot_entered)
    {
      _boot_entered = _tryEnterToBoot();
      _last_new_data_ts = systemMillis();
    }
    else
    {
      if(_temp_buffer != NULL)
      {
        _tryProcessBufferData();
      }
      else
      {
        _last_handle_result = HandleResult::NO_NEW_DATA;
      }
    }

    if(_last_handle_result == HandleResult::NO_NEW_DATA)
    {
      if(systemMillis() - _last_new_data_ts > NO_NEW_DATA_TIMEOUT_MS)
      {
        _setFatalResult(NO_NEW_DATA_TIMEOUT);
      }
    }
  }
  return _last_handle_result;
}

}
