/*
Copyright (C) 2017 Nanopix
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
* Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in
the documentation and/or other materials provided with the
distribution.
* Neither the name of Nanopix nor the names
of its contributors may be used to endorse or promote products
derived from this software without specific prior written
permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/


#include "UdpPacketSender.h"
#include "nanopix_config.h"

namespace NanopixCore {

/* =========================================== */
/* =============== Constructors ============== */
/* =========================================== */

UdpPacketSender::UdpPacketSender()
{
    ip = IPAddress(192, 168, 43, 140);
    gateway = IPAddress(192, 168, 43, 1);
    subnet = IPAddress(255, 255, 255, 0);
    multicastIP = IPAddress(192, 168, 43, 255);

    for (int i = 0; i < controllersCount; i++)
        controllerIP[i] = IPAddress(192, 168, 43, 255);
}

int UdpPacketSender::initialize(AsyncUDP * udp)
{
    int ret = NANOPIX_STATUS_OK;
    if (udp == NULL)
        return NANOPIX_STATUS_NULL_ARGUMENT;

    this->udp = udp;

    return ret;
}

/* =========================================== */
/* ================ Operations =============== */
/* =========================================== */

int UdpPacketSender::sendToConnected(uint8_t * data, int size, SenderSourceType sourceType) const
{
    int ret = NANOPIX_STATUS_OK;
    if(data == NULL)
        return NANOPIX_STATUS_NULL_ARGUMENT;
    if(size <= 0)
        return NANOPIX_STATUS_INVALID_ARGUMENT;

    if(controllersCount == 0)
        return NANOPIX_STATUS_NO_DATA;

    int port = resolvePortBySourceType(sourceType);
    for(int i = 0; i < controllersCount; i++) 
    {
        udp->writeTo(data, size, controllerIP[i], port);
    }

    return ret;
}

int UdpPacketSender::sendMulticast(uint8_t * data, int size, SenderSourceType sourceType) const
{
    int ret = NANOPIX_STATUS_OK;
    if (data == NULL)
        return NANOPIX_STATUS_NULL_ARGUMENT;
    if (size <= 0)
        return NANOPIX_STATUS_INVALID_ARGUMENT;

    int port = resolvePortBySourceType(sourceType);
    udp->writeTo(data, size, multicastIP, port);

    return ret;
}

int UdpPacketSender::registerController(IPAddress addr) 
{
    int ret = NANOPIX_STATUS_OK;

    if(controllersCount == MAX_CONTROLLERS)
        return NANOPIX_STATUS_FULL;

    controllerIP[controllersCount] = addr;
    controllersCount++;

    return ret;
}

/* =========================================== */
/* ================ Accessors ================ */
/* =========================================== */

int UdpPacketSender::connectedCount() const
{
    return controllersCount;
}

bool UdpPacketSender::hasController(IPAddress addr) const
{
    for (int i = 0; i < controllersCount; i++)
        if (addr == controllerIP[i])
            return true;
    return false;
}

bool UdpPacketSender::isMainController(IPAddress addr) const
{
  return ((controllersCount > 0) && (addr == controllerIP[0]));
}

int UdpPacketSender::resolvePortBySourceType(SenderSourceType sourceType) const
{
    switch (sourceType)
    {
        case SenderSourceType::ResendFromFC:
            return CONTROLLER_UDP_PORT;
        break;
        case SenderSourceType::OwnMessage:
            return ESP_LISTENER_UDP_PORT;
        break;
        default:
            return 0;
    }
}

}
