/*
    Copyright (C) 2017 Nanopix
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the 
      distribution.
    * Neither the name of Nanopix nor the names
      of its contributors may be used to endorse or promote products
      derived from this software without specific prior written
      permission.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
    SUCH DAMAGE.
*/

#include "settings_manager.h"

namespace NanopixCore {
	
SettingsManager * SettingsManager::instance = NULL;

SettingsManager::SettingsManager() {
  _udpStarted = false;
	EEPROM.begin(EEPROM_SIZE_BYTES);
}

SettingsManager & SettingsManager::getInstance()
{
	if(instance == NULL)
	{
		instance = new SettingsManager();
	}
	return *instance;
}

SettingsManager::~SettingsManager() {
	if(instance != NULL)
	{
		delete instance;
	}
	instance = NULL;
}

void SettingsManager::initAll(UdpPacketSender * udpSender)
{
  this->udpSender = udpSender;
  
  randomSeed(ESP.getCycleCount());
  if(isEepromInitialized())
  {
	  loadSSID_Pass(SSID, PASS);
	  loadWifiChannel();
  } 
  else 
  {
	  initEeprom();
	  resetSSID_PassToDefault();
	  wifiChannel = 0; // if zero - then search optimal
  }
}

void SettingsManager::startUdp()
{
  if(!_udpStarted)
  {
    espUDP.begin(localESP_UdpPort);
    _udpStarted = true;
  }
}

void SettingsManager::parseMavPackets()
{
  int packetSize = espUDP.parsePacket();
  if (packetSize) {
	int len = espUDP.read(buffer, 256);
	for(int i = 0; i < len; i++) {
	  uint8_t res = mavlink_parse_char(1, buffer[i], &rxMessage, &mavStatus);
	  if(res)
	  {
		 uint8_t packID = rxMessage.msgid;
		 String msgText = "";
		 switch(packID) {
		  case MAVLINK_MSG_ID_STATUSTEXT:
			mavlink_msg_statustext_decode(&rxMessage, &statusTextMsg);
			msgText = String(statusTextMsg.text);

			if(msgText == String("ESP:ssid_pass_request"))
			{
			  sendSSID_PASS();
			  sendWifiChannel();
			}

			if(msgText.substring(0, 12) == String("ESP:setAuth:"))
			{
/*#if DEBUG
			  Serial.printf("ESP:setAuth command received\n");
#endif*/
			  int tokenPos;
			  SSID = substringToToken(msgText, 12, ',', tokenPos);
			  PASS = substringToToken(msgText, tokenPos+1, ';', tokenPos);
			  
			  sendSSID_PASS();
			  saveSSID_Pass(SSID, PASS);
			}
			if(msgText.substring(0, 12) == String("ESP:setChan:"))
			{
/*#if DEBUG
			  Serial.printf("ESP:setChan command received\n");
#endif*/
			  int newChan = 0;
			  newChan = atoi(msgText.substring(12, 14).c_str());
			  if(newChan >= 0 && newChan <= 12)
			  {
				 char buf[64];
				 sprintf(buf, "WiFi channel set to %d", newChan);
				 sendString(buf);
				 wifiChannel = newChan;
				 saveWifiChannel();
			  }
			}
			if(msgText.substring(0, 10) == String("ESP:reboot"))
			{
/*#if DEBUG
			  Serial.printf("Reboot command received\n");
#endif*/
			  ESP.restart();
			}
			if(msgText.substring(0, 17) == String("ESP:setRandomSSID"))
			{
/*#if DEBUG
			  Serial.printf("ESP:setRandomSSID command received\n");
#endif*/
			  resetSSID_PassToDefault();
			  sendSSID_PASS();
			  wifiChannel = 0;
			  saveWifiChannel();
			}
/*#if DEBUG                
			sendString(msgText.c_str());
#endif*/
		  break;
		  
		  case MAVLINK_MSG_ID_PARAM_REQUEST_READ:
			mavlink_param_request_read_t paramRequestReadMsg;
			mavlink_msg_param_request_read_decode(&rxMessage, &paramRequestReadMsg);
			sendParam(paramRequestReadMsg.param_id);
		  break;
		  
		  default:
		  
		  break;
		 }
	  }
	}
  }
}

bool SettingsManager::sendParam(char* strId)
{
  int paramsCount = 2;
  if(strcmp(strId, ESP_SOFTWARE_VERSION_ID) == 0)
  {
	int paramIndex = 0;
	mavlink_message_t txMsg;
	mavlink_msg_param_value_pack(0, 0, &txMsg, ESP_SOFTWARE_VERSION_ID, ESP_SOFTWARE_VERSION, MAVLINK_TYPE_FLOAT, paramsCount, paramIndex);
	int txSize = mavlink_msg_to_send_buffer(buffer, &txMsg);

	udpSender->sendToConnected(buffer, txSize, SenderSourceType::OwnMessage);
	return true;
  }
  if(strcmp(strId, ESP_WIFI_CHANNEL_ID) == 0)
  {
	int paramIndex = 1;
	mavlink_message_t txMsg;
	mavlink_msg_param_value_pack(0, 0, &txMsg, ESP_WIFI_CHANNEL_ID, wifiChannel, MAVLINK_TYPE_FLOAT, paramsCount, paramIndex);
	int txSize = mavlink_msg_to_send_buffer(buffer, &txMsg);

	udpSender->sendToConnected(buffer, txSize, SenderSourceType::OwnMessage);
	return true;
  }
  return false;
}

void SettingsManager::resetSSID_PassToDefault()
{
  generateRandomSSID(SSID);
  PASS = DEFAULT_PASSWORD;
  saveSSID_Pass(SSID, PASS);
/*#if DEBUG
  Serial.printf("SSID and pass random generated\n");
#endif*/
}

bool SettingsManager::isDefaultPasswordNow()
{
  return PASS == String(DEFAULT_PASSWORD);
}

void SettingsManager::sendSSID_PASS()
{
  sendString( ("ESP.auth:"+SSID + ":" + PASS).c_str());
}

void SettingsManager::sendWifiChannel()
{
  char buf[64];
  sprintf(buf, "ESP.wifiChan:%d", wifiChannel);
  sendString(buf);
}

String SettingsManager::substringToToken(String src, int start, char token, int& tokenPos)
{
  char buf[50];
  int i = 0;
  for(; i < 50; i++)
  {
	if(src[start+i] == token)
	  break;
	buf[i] = src[start+i];
  }
  tokenPos = start+i;
  buf[i] = 0;
  return String(buf);
}

void SettingsManager::sendString(const char* str)
{
  mavlink_msg_statustext_pack(0,0,&txMessage, 0, str);
  int16_t txSize = mavlink_msg_to_send_buffer(buffer, &txMessage);
  udpSender->sendToConnected(buffer, txSize, SenderSourceType::OwnMessage);
}

void SettingsManager::sendString(String str)
{
  sendString(str.c_str());
}

uint8_t SettingsManager::getWifiChannel()
{
  return wifiChannel;
}

void SettingsManager::loadWifiChannel()
{
  wifiChannel = EEPROM.read(WIFI_CHANNEL_ADDR);
  if(wifiChannel > 12 || wifiChannel < 0)
	wifiChannel = 0;
}

void SettingsManager::saveWifiChannel()
{
  EEPROM.write(WIFI_CHANNEL_ADDR, wifiChannel);
  EEPROM.commit();
}

SettingsManager::BOOT_MODES SettingsManager::getNextBootMode()
{
  BOOT_MODES bootMode = (BOOT_MODES)(EEPROM.read(BOOT_MODE_ADDR));
  if((bootMode < BOOT_MODES::First) || (bootMode > BOOT_MODES::Last))
  {
  	bootMode = TRY_ESPNOW_BOOT_MODE;
  }
  return bootMode;
}

void SettingsManager::setNextBootMode(SettingsManager::BOOT_MODES bootMode)
{
  EEPROM.write(BOOT_MODE_ADDR, (uint8_t)bootMode);
  EEPROM.commit();
}

bool SettingsManager::isEepromInitialized()
{
  byte checkByte1 = EEPROM.read(0);
  byte checkByte2 = EEPROM.read(1);
  bool res = ((checkByte1 == START_TOKEN_BYTE1) && (checkByte2 == START_TOKEN_BYTE2));
/*#if DEBUG
  Serial.printf("EEPROM wrong tokens: %d, %d | target: %d, %d\n", checkByte1, checkByte2, START_TOKEN_BYTE1, START_TOKEN_BYTE2);
#endif*/
  return res;
}

void SettingsManager::initEeprom()
{
/*#if DEBUG
  Serial.printf("init EEPROM tokens\n");
#endif*/
  EEPROM.write(0, START_TOKEN_BYTE1);
  EEPROM.write(1, START_TOKEN_BYTE2);
  for(int i = 2; i < EEPROM_SIZE_BYTES; ++i)
  {
    EEPROM.write(i, 0);
  }
  EEPROM.commit();
}

void SettingsManager::generateRandomSSID(String& ssid)
{
  int randNumber = random(100000);
  String num = String(randNumber);
  ssid = "Nanopix" + num;
/*#if DEBUG
  Serial.printf("Random SSID generated: %s\n", ssid.c_str());
#endif*/
}

void SettingsManager::saveSSID_Pass(String ssid, String pass)
{
  int i = AUTH_DATA_ADDR;
  for(int k = 0; ssid[k] != 0; k++)
	EEPROM.write(i++, ssid[k]);
  EEPROM.write(i++, 0);
  for(int k = 0; pass[k] != 0; k++)
	EEPROM.write(i++, pass[k]);
  EEPROM.write(i++, 0);
  EEPROM.commit();
}

void SettingsManager::loadSSID_Pass(String& ssid, String& pass)
{
  char ssidBuf[MAX_SSID_BYTES], passBuf[MAX_PASS_BYTES];
  byte readByte = 1;
  int addr = AUTH_DATA_ADDR;
  int k = 0;
  do {
	ssidBuf[k] = readByte = EEPROM.read(addr);
	addr++;
	k++;
  } while(readByte != 0);
  k = 0;
  do {
	passBuf[k] = readByte = EEPROM.read(addr);
	addr++;
	k++;
  } while(readByte != 0);
  ssid = String(ssidBuf);
  pass = String(passBuf);
}
					


}

