/*
Copyright (C) 2018 Nanopix
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
* Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in
the documentation and/or other materials provided with the
distribution.
* Neither the name of Nanopix nor the names
of its contributors may be used to endorse or promote products
derived from this software without specific prior written
permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

#include "stm_ota_udp_receiver.h"
#include "STM32_OTA.h"
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>

namespace NanopixCore
{

static WiFiUDP udp_ota;
static IPAddress remote_ip_addr;



bool STM_OTA_UDP_Receiver::start(int udp_local_port, int udp_remote_port)
{
    udp_ota.begin(udp_local_port);
    STM32_OTA::getInstance().begin();
    _udp_local_port = udp_local_port;
    _udp_remote_port = udp_remote_port;
    _current_state = UploadFlash;
    lastPacketTimestamp = 0;
}

bool STM_OTA_UDP_Receiver::restart()
{
    udp_ota.flush();
    udp_ota.stop();
    start(_udp_local_port, _udp_remote_port);
}

bool STM_OTA_UDP_Receiver::handle()
{
    while (true)
    {
        int packet_size = udp_ota.parsePacket();
        if (packet_size)
        {
            lastPacketTimestamp = millis();
            int len = udp_ota.read(_rx_buffer, RX_BUFFER_SIZE);
            if (len == 1) // then it is command
            {
                remote_ip_addr = udp_ota.remoteIP();
                STM32_OTA::BootloaderAckStatus res;
                uint8_t command = _rx_buffer[0];
                switch (command)
                {
                case EnterBOOT:
                    res = STM32_OTA::getInstance().enterToBOOT();
                    sendLog(res == STM32_OTA::BOOTLOADER_ACK_OK ? 
                        "Enter to BOOT mode - OK" : 
                        "Enter to BOOT mode - FAILURE!");
                    break;
                case StartUpload:
                    _current_state = UploadFlash;
                    STM32_OTA::getInstance().startUpdate();
                    sendLog("STM OTA: start flash...");
                    break;
                case StartVerify:
                    _current_state = VerifyFlash;
                    STM32_OTA::getInstance().startUpdate();
                    sendLog("STM OTA: start verify...");
                    break;
                case GlobalErase:
                    res = STM32_OTA::getInstance().globalEraseEx();
                    if (res != STM32_OTA::BOOTLOADER_ACK_OK)
                    {
                        res = STM32_OTA::getInstance().globalErase();
                        sendLog(res == STM32_OTA::BOOTLOADER_ACK_OK ? 
                            "Extended global erase FAIL, standart OK" : 
                            "Extended and standart global erase both FAILURE");
                    }
                    else
                    { 
                        sendLog("Extended global erase OK");
                    }
                    break;
                case ReadoutUnprotect:
                    res = STM32_OTA::getInstance().setReadoutUnprotect();
                    sendLog(res == STM32_OTA::BOOTLOADER_ACK_OK ? 
                        "Readout unprotect OK" : 
                        "Readout unprotect FAILURE");
                    break;
                case WriteUnprotect:
                    res = STM32_OTA::getInstance().setWriteUnprotect();
                    sendLog(res == STM32_OTA::BOOTLOADER_ACK_OK ? 
                        "Write unprotect OK" : 
                        "Write unprotect FAILURE");
                    break;
                case FinishUpdate:
                    STM32_OTA::getInstance().finishUpdate();
                    lastPacketTimestamp = 0;
                    return true;
                default:
                    break;
                }

            }
            else 
            {
                ////////////////////////////////
                // TODO: need check packet CRC!!!!!!
                ////////////////////////////////
                size_t completed_bytes = 0;
                if (_rx_buffer[3] != 0) // then it is address settings packet
                {
                    STM32_OTA::getInstance().write(_rx_buffer, RX_BUFFER_SIZE, completed_bytes);
                    sendResponce(STM32_OTA::ACK_TOKEN_BYTE); // TODO: move to stm32 ota
                }
                else
                {
                    switch (_current_state)
                    {
                    case UploadFlash:
                        
                        if (STM32_OTA::getInstance().write(_rx_buffer, RX_BUFFER_SIZE, completed_bytes) == STM32_OTA::BOOTLOADER_ACK_OK)
                        {
                            sendResponce(STM32_OTA::ACK_TOKEN_BYTE);
                            sendResponce(STM32_OTA::ACK_TOKEN_BYTE);
                        }
                        break;
                    case VerifyFlash:
                        if (STM32_OTA::getInstance().verify(_rx_buffer) == STM32_OTA::BOOTLOADER_ACK_OK)
                            sendResponce(STM32_OTA::ACK_TOKEN_BYTE);
                        break;
                    }
                }
            }
        }
        else 
        {
            if (STM32_OTA::getInstance().isOTA_Updating() == false)
                break;
            if(millis() - lastPacketTimestamp > RX_PACKET_TIMEOUT_MS)
                break;
        }
    } // while true end
    STM32_OTA::getInstance().finishUpdate();
    lastPacketTimestamp = 0;
    return false;
}

bool STM_OTA_UDP_Receiver::isOTA_Request()
{
    return udp_ota.available();
}

void STM_OTA_UDP_Receiver::sendLog(const char* str)
{
    udp_ota.beginPacket(remote_ip_addr, _udp_remote_port);
    udp_ota.write(str);
    udp_ota.endPacket();
}

void STM_OTA_UDP_Receiver::sendResponce(uint8_t val)
{
    udp_ota.beginPacket(remote_ip_addr, _udp_remote_port);
    String hex_string = String(val, HEX);
    char char_var[sizeof(hex_string)];
    hex_string.toCharArray(char_var, sizeof(char_var));
    udp_ota.write(char_var);
    udp_ota.endPacket();
}

STM_OTA_UDP_Receiver::UpdateStates STM_OTA_UDP_Receiver::_current_state;
uint8_t STM_OTA_UDP_Receiver::_rx_buffer[RX_BUFFER_SIZE];
int STM_OTA_UDP_Receiver::_udp_remote_port;
int STM_OTA_UDP_Receiver::_udp_local_port;
unsigned long STM_OTA_UDP_Receiver::lastPacketTimestamp;

}
